#include <kos.h>
#include <ctype.h>
#include "driver.h"
#include "ticker.h"
#define WT_PROFILE_GLOBALS
#include "dreamcast.h"
#include "config.h"
#include "rc.h"
#include <png/png.h>

/* ----------------------------------------------------- */
/* Eventually this lot will come from a template file... */
/* ----------------------------------------------------- */

#define BACKDROP_NAME    "BACKDROP"
#define BACKDROP_FILTER  PVR_FILTER_NONE
#define BACKDROP_X       0
#define BACKDROP_Y       0
#define BACKDROP_Z       1
#define BACKDROP_WIDTH   640
#define BACKDROP_HEIGHT  480

#define SNAPSHOT_ENABLED 0
#define SNAPSHOT_FILTER  PVR_FILTER_BILINEAR
#define SNAPSHOT_X       12
#define SNAPSHOT_Y       12
#define SNAPSHOT_Z       2
#define SNAPSHOT_WIDTH   316
#define SNAPSHOT_HEIGHT  236

#define GUI_FONT         "wt_font5"                             
                             
#define MENU_RED         0
#define MENU_GREEN       0
#define MENU_BLUE        0
#define MENU_FONT        0.5f                            
#define MENU_X           366
#define MENU_Y           16
#define MENU_Z           2
#define MENU_WIDTH       258
#define MENU_HEIGHT      288

#define OPTIONS_RED      0.3f
#define OPTIONS_GREEN    0.1f
#define OPTIONS_BLUE     0.1f
#define OPTIONS_FONT     0.5f
#define OPTIONS_X        366
#define OPTIONS_Y        346
#define OPTIONS_Z        2
#define OPTIONS_WIDTH    258
#define OPTIONS_HEIGHT   118
      
#define INFO_RED         0.1f
#define INFO_GREEN       0.1f
#define INFO_BLUE        0.3f
#define INFO_FAIL_RED    0.3f
#define INFO_FAIL_GREEN  0.1f
#define INFO_FAIL_BLUE   0.3f
#define INFO_FONT        0.5f
#define INFO_X           16
#define INFO_Y           285
#define INFO_Z           2
#define INFO_WIDTH       258
#define INFO_HEIGHT      179                            
                            
#define NOROM_RED        1
#define NOROM_GREEN      1
#define NOROM_BLUE       1
#define NOROM_FONT       0.7f
#define NOROM_X          10
#define NOROM_Y          10
#define NOROM_Z          2

//KOS_INIT_FLAGS(INIT_DEFAULT|INIT_QUIET);
KOS_INIT_FLAGS(INIT_DEFAULT);

#if defined(WT_ROMDISK)
extern uint8 romdisk_boot[];
KOS_INIT_ROMDISK(romdisk_boot);
#define font_dir    "/rd"
#define sample_path "/rd/samples;/cd/samples"
#define rom_path    "/rd/roms;/cd/maim;/cd/roms"
#else
#define font_dir    "/cd/font"
#define rom_path    "/cd/maim;/cd/roms"
#define sample_path "/cd/samples"
#endif

unsigned long dbgvalue = 0;

int errno = 0;

/* fileio.c */ extern struct rc_option fileio_opts[];
/* fileio.c */ extern const char *get_path_for_filetype(int filetype, int pathindex, UINT32 *count);
/* fileio.c */ extern void set_pathlist(int file_type, const char *new_rawpath);

/* input.c */ extern maple_device_t *maple_keyboard_device;
/* input.c */ extern maple_device_t *maple_controller_dev[MAPLE_PORT_COUNT+1];
/* input.c */ extern unsigned char   controller_type[MAPLE_PORT_COUNT];

/* input.c */ extern kbd_state_t def_kbd_state;
/* input.c */ extern kbd_state_t *kbd_state;

/* input.c */ extern cont_state_t def_cont_state;
/* input.c */ extern cont_state_t *cont_state[MAPLE_PORT_COUNT];

/* input.c */ extern unsigned char max_players;
/* input.c */ extern unsigned char ui_mode;          /* 0=CONSOLE, 1=ARCADE */

/* input.c */ extern unsigned long js_calls;
/* video.c */ extern unsigned long md_calls;

/* video.c */ extern unsigned char use_video_thread; /* 0=NO, 1=YES */
/* video.c */ extern unsigned char use_display_mode; /* 0=STRETCH, 1=1x, 2=2x, 3=STRETCH (No filter) */ 
/* video.c */ extern unsigned char use_throttle;     /* 0=NO, 1=YES */
/* video.c */ extern unsigned char use_dma;          /* 0=NO, 1=YES */
/* video.c */ extern int           frameskip;
/* video.c */ extern int           autoframeskip;
/* video.c */ extern float         avg_fps; 
/* video.c */ extern int           showfps;
/* video.c */ extern unsigned char screen_orientation;

/* sound.c */ extern unsigned char use_audio_throttle;   /* 0=NO, 1=YES */
/* sound.c */ extern unsigned char use_audio_thread;     /* 0=YES MID, 1=NO MID, 2=INTELLIGENT, 3=NO END, 4=YES END */
/* sound.c */ extern unsigned char use_audio_force_mono; /* 0=NO, 1=YES */

/* cpuexec.c */ extern double cpu_speed[3];

maple_device_t *first_controller;
cont_state_t *first_cont_state = 0;

int play_sound = 0;
int cpu_slices_per_frame_override = 0;
int gResolution = 0;

extern unsigned long kp_count;

/* .-----------------------------------------------------------------------------. */
/* | AUDIO               | VIDEO                                                 | */
/* |---------------------|-------------------------------------------------------| */
/* | FREQ | THRD | ATHRT | DMDE | SFPS | RES  | XMDE | THRD | COLM | THRT | FSKP | */
/* '-----------------------------------------------------------------------------' */

int global_dc_failure = 0;
int global_colormode = 0;
int global_frameskip = 0;

char *maim_option_name[MAIM_OPTIONS]    = {"Audio Frequency:", 
                                           "Audio Thread:", 
                                           "Audio Throttle:",
                                           "Display Mode:",
                                           "Show FPS:",
                                           "CPU Slices:",
                                           "Video DMA:",
                                           "Video Thread:",
                                           "Colour Mode:",
                                           "Video Throttle:",
                                           "Frameskip Level:",
                                           "Screen Orientation:",
                                           "CPU0 Speed:",
                                           "CPU1 Speed:",
                                           "CPU2 Speed:",
                                           "Other Value"};

int maim_option_cur[MAIM_OPTIONS] = TINY_OPTIONS;
int maim_option_def[MAIM_OPTIONS] = TINY_OPTIONS;
int maim_option_inc[MAIM_OPTIONS] = {4410,  1, 1, 1, 1, 1,   1, 1, 1, 1, 1,  1, 1,   1,   1,   100};
int maim_option_max[MAIM_OPTIONS] = {44100, 4, 1, 3, 1, 200, 1, 1, 4, 1, 12, 3, 200, 200, 200, 5000};

//#define CHUNK_BUFFER_SIZE 14417920   // 13.75MB
//#define CHUNK_BUFFER_SIZE 14155776   // 13.5MB
//#define CHUNK_BUFFER_SIZE 13631488   //13MB (13x1048576)
//#define CHUNK_BUFFER_SIZE 14680064   //14MB (14x1048576)
//#define CHUNK_BUFFER_SIZE 13094400   //12.5MB
#define CHUNK_BUFFER_SIZE 10485760

#if defined(WT_MALLOCDBG)

#undef malloc
#undef memalign
#undef realloc
#undef free

void *osd_malloc(int length,char *filename, int line)
{
  void *v=malloc(length);

  printf("malloc: %p (%d) from %s line %d\n",v,length,filename,line);

  return(v);
}

void *osd_memalign(int alignment, int length, char *filename, int line)
{
  void *v=memalign(alignment,length);

  printf("memalign: %p (%d,%d) from %s line %d\n",v,alignment,length,filename,line);

  return(v);
}

void *osd_realloc(void *ov,int length,char *filename, int line)
{
  void *v=realloc(ov,length);

  printf("realloc: %p->%p (%d) from %s line %d\n",ov,v,length,filename,line);

  return(v);
}

void osd_free(void *v, char *filename, int line)
{
  printf("free: %p from %s line %d\n",v,filename,line);

  return(free(v));
}
#endif

/* osd_init ===================================================== */
/*                                                                */
/* Initialise Controllers / Keyboard.                             */
/*                                                                */
/* ============================================================== */

int osd_init(void)
{
  int port = 0;
  int unit = 0;
  int ctot = 0;

  /* ---------------------------------------- */
	/* Determine which controlers are available */
  /* ---------------------------------------- */

  /* Should check here to see if we've already got controllers */

  first_controller = 0;

  memset(&def_kbd_state,0,sizeof(kbd_state_t));
  memset(&def_cont_state,0,sizeof(cont_state_t));
  memset(&maple_controller_dev,0,sizeof(maple_device_t *)*(MAPLE_PORT_COUNT+1));

  max_players = 0;

	for (port=0; port<4; port++)
  {
	  for (unit=0; unit<6; unit++) 
		{
		  if ((!max_players || ctot < max_players) && maple_device_func(port, unit) & (MAPLE_FUNC_CONTROLLER | MAPLE_FUNC_MOUSE))
      {
        maple_controller_dev[ctot] = maple_enum_dev(port,unit);

        if (maple_device_func(port, unit) & MAPLE_FUNC_MOUSE)
          controller_type[ctot] = 8;
        else if (maple_device_func(port, unit) & MAPLE_FUNC_LIGHTGUN)
          controller_type[ctot] = 4;
        else if (!strncmp(maple_controller_dev[ctot]->info.product_name,"Racing",6))
          controller_type[ctot] = 2;
        else
        {
          controller_type[ctot] = 1;

          if (!first_controller)
            first_controller = maple_controller_dev[ctot];
        }

        cont_state[ctot] = &def_cont_state;

        ctot++;
      }
      else if (!maple_keyboard_device && maple_device_func(port,unit) & MAPLE_FUNC_KEYBOARD)
      {
        maple_keyboard_device = maple_enum_dev(port,unit);

        kbd_state = &def_kbd_state;
      }
		}
	}

  /* ------------------------------------------------------------------------ */
  /* Take the number of controllers found to be the maximum number of players */
  /* ------------------------------------------------------------------------ */

  max_players = ctot;

	return 0;
}

/* osd_exit ===================================================== */
/*                                                                */
/* Shut down any audio stream that was started.                   */
/*                                                                */
/* ============================================================== */

void osd_exit(void)
{
  //term_audio_stream();
} 

/* font_load ==================================================== */
/*                                                                */
/* ============================================================== */

typedef struct wt_font
{
  int            *pos_array;
  pvr_poly_hdr_t  hdr_array[2];
  pvr_ptr_t       txr_array[2];
  int             height;
  int             txr_height;
  int             txr_width;
} 
wt_font_t;

void splash_screen(wt_font_t *font, uint8 in);

uint8 font_load(char *font_name, wt_font_t *font)
{
  int loaded = 1;

  if (font_name && font)
  {
    FILE *fp = 0;

    unsigned char name[256];

    sprintf(name,"%s/%s.wtf",font_dir,font_name);

    if ((fp = fopen(name,"rb")))
    {
      int n;
      int num_files = 0;
      
      memset(font,0x00,sizeof(wt_font_t));

      fread(&font->txr_width,sizeof(font->txr_width),1,fp);
      fread(&font->txr_height,sizeof(font->txr_height),1,fp);
      fread(&num_files,sizeof(num_files),1,fp);
      fread(&font->height,sizeof(font->height),1,fp);

      for(n=0;loaded && n<num_files;n++)
      {
        if ((font->txr_array[n] = pvr_mem_malloc(font->txr_width*font->txr_height*2)))
        {
          pvr_poly_cxt_t cxt;

          unsigned char font_file[200];

          int i = 0;

          fread(&font_file[i],sizeof(unsigned char),1,fp);

          while(font_file[i++])
            fread(&font_file[i],sizeof(unsigned char),1,fp);

          sprintf(name,"%s/%s",font_dir,font_file);

          printf("%s %d\n",__FILE__,__LINE__);

          png_to_texture(name,font->txr_array[n],PNG_FULL_ALPHA);

          printf("%s %d\n",__FILE__,__LINE__);

          pvr_poly_cxt_txr(&cxt, PVR_LIST_TR_POLY, PVR_TXRFMT_ARGB4444, 1024, 32, font->txr_array[n], PVR_FILTER_BILINEAR);
          pvr_poly_compile(&font->hdr_array[n], &cxt);

        }
        else
          loaded = 0;
      }

      if (loaded && (font->pos_array = malloc(256*sizeof(int))))
        fread(font->pos_array,sizeof(int),256,fp);
      else
        loaded = 0;

      fclose(fp);
    }
    else
      loaded = 0;
  }
  else
    loaded = 0;

  return(loaded);
}

/* font_unload ================================================== */
/*                                                                */
/* ============================================================== */

void font_unload(wt_font_t *font)
{
  if (font)
  {
    int n=0;

    free(font->pos_array);

    while(n<2 && font->txr_array[n])
      pvr_mem_free(font->txr_array[n++]);
  }

  return;
}

/* font_write =================================================== */
/*                                                                */
/* ============================================================== */

#define MAX(A,B) ((A)>(B)?(A):(B))

void font_write(wt_font_t *font, unsigned char *string, float mult, 
                int x, int y, int max_x, float z, 
                float fade, float red, float green, float blue,
                int *wid_out, int *hgt_out, uint8 display)
{
  int   n       = 0;
  int   cur_x   = x;
  int   cur_y   = y;
  int   max_wid = 0;
  int   hgt     = (int)(((float)font->height)*mult);
  int   max_hgt = hgt;

  int   wid;
  float u1,u2;
  
  while(string[n])
  {
    switch(string[n])
    {
      case '\n':
        max_wid = MAX((cur_x-x),max_wid);
        cur_x   =  x;
        cur_y   += max_hgt;
        max_hgt = hgt;
        break;

      case '\1':            /* size inc */
        if (string[n+1])
        {
          mult = mult * ((float)string[++n]);

          if ((hgt  = (int)(((float)font->height)*mult)) > max_hgt)
            max_hgt = hgt;
        }
        break;

      case '\2':            /* size dec */
        if (string[n+1])
        {
          if ((mult = mult / ((float)string[++n])) < 0)
            mult = 0.0;

          hgt = (int)(((float)font->height)*mult);
        }
        break;

      case '\3':            /* x inc */
        if (string[n+1])
        {
          int delta = (int)(((float)string[++n])*mult);

          if (!max_x || (cur_x + delta < max_x))
            cur_x += delta;
        }
        break;

      case '\4':            /* x dec */
        if (string[n+1])
          cur_x -= (int)(((float)string[++n])*mult);;
        break;

      case '\5':            /* y inc */
        if (string[n+1])
          cur_y += (int)(((float)string[++n])*mult);
        break;

      case '\6':            /* y dec */
        if (string[n+1])
          cur_y -= (int)(((float)string[++n])*mult);
        break;

      default:
        {
          int bef = font->pos_array[string[n]-1];
          int aft = font->pos_array[string[n]];

          wid = (int)((float)(aft-bef))*mult;


          if (wid>0)
          {
            if (aft>1024)
            {
              if (bef<1024)
                bef = 0;
              else
                bef -= 1024;

              aft -= 1024;

              wid = (int)((float)(aft-bef))*mult;
            }

            /* --------------------------------- */
            /* we've written the maximum allowed */
            /* --------------------------------- */

            if (max_x && (cur_x + wid > max_x))
              break;

            if (display)
            {
              pvr_vertex_t vert; 

              u1 = ((float)bef) / 1024.0f;
              u2 = ((float)aft) / 1024.0f;

              pvr_prim(&font->hdr_array[font->pos_array[string[n]]/1024], sizeof(pvr_poly_hdr_t)); 

              vert.flags = PVR_CMD_VERTEX;
              vert.x = cur_x;
              vert.y = cur_y;
              vert.z = z;
              vert.u = u1;
              vert.v = 0;
              vert.argb = PVR_PACK_COLOR (fade, red, green, blue);

              vert.oargb = 0;
              pvr_prim(&vert, sizeof(pvr_vertex_t));

              vert.x = cur_x+wid;
              vert.u = u2;
              pvr_prim(&vert,sizeof(pvr_vertex_t));

              vert.x = cur_x;
              vert.y = cur_y+hgt;
              vert.u = u1;
              vert.v = 1;
              pvr_prim(&vert,sizeof(pvr_vertex_t));

              vert.flags = PVR_CMD_VERTEX_EOL;
              vert.x = cur_x+wid;
              vert.u = u2;
              pvr_prim(&vert,sizeof(pvr_vertex_t));
            }

            cur_x+=wid;
          }
        }
        break;
    }

    n++;
  }

  if (wid_out)
    *wid_out = MAX((cur_x-x),max_wid);

  if (hgt_out)
    *hgt_out = (cur_y+max_hgt)-y;

  return;
}

pvr_ptr_t      txr_snap = NULL;
pvr_poly_hdr_t hdr_snap;
pvr_ptr_t      txr_back = NULL;
pvr_poly_hdr_t hdr_back;

/* gfx_init ===================================================== */
/*                                                                */
/* ============================================================== */

void gfx_init(void)
{
  int pv = 0;

  pvr_poly_cxt_t cxt;

  if ((txr_snap = pvr_mem_malloc(256*256*2)))
  {
    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565, 256, 256, txr_snap, SNAPSHOT_FILTER);
    pvr_poly_compile(&hdr_snap, &cxt);
  }

  if ((pv=mame_faccess(BACKDROP_NAME,FILETYPE_SCREENSHOT)) && (txr_back = pvr_mem_malloc(1024*512*2)))
  {
    char name[256];

    sprintf(name,"%s/%s.png",get_path_for_filetype(FILETYPE_SCREENSHOT,pv-1,0),BACKDROP_NAME);

    printf("%s %d\n",__FILE__,__LINE__);

    png_to_texture(name,txr_back,PNG_NO_ALPHA);

    printf("%s %d\n",__FILE__,__LINE__);

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565, 1024, 512, txr_back, BACKDROP_FILTER);
    pvr_poly_compile(&hdr_back, &cxt); 

  }
  else
    txr_back = NULL;
}

/* gfx_shutdown ================================================= */
/*                                                                */
/* ============================================================== */

void gfx_shutdown(void)
{
  pvr_mem_free(txr_back);
  pvr_mem_free(txr_snap);
}

/* gui_init ===================================================== */
/*                                                                */
/* ============================================================== */

int gui_init(wt_font_t *font, int pvr_init_needed)
{
  pvr_init_params_t gui_params = 
  {
	  { 
      PVR_BINSIZE_16,
      PVR_BINSIZE_0,
      PVR_BINSIZE_16,
      PVR_BINSIZE_0,
      PVR_BINSIZE_0
    },

	  /* Vertex buffer size 512K */
	  512*1024
  };

  if (pvr_init_needed)
	  pvr_init(&gui_params);

  pvr_set_bg_color(255,0,0);

  gfx_init();

  if (font_load(GUI_FONT,font))
  {
    //splash_screen(font,0);

    return(1);
  }

  return(0);
}

/* gui_shutdown ================================================= */
/*                                                                */
/* ============================================================== */

void gui_shutdown(wt_font_t *font, int pvr_shutdown_needed)
{
  splash_screen(font,1);

  font_unload(font);

  gfx_shutdown();

  if (pvr_shutdown_needed)
    pvr_shutdown();
}

/* gui_begin ==================================================== */
/*                                                                */
/* ============================================================== */

void gui_begin()
{
  pvr_wait_ready();
  pvr_scene_begin();
}

/* gui_gfx_begin ================================================ */
/*                                                                */
/* ============================================================== */

void gui_gfx_begin(void)
{
  pvr_list_begin(PVR_LIST_OP_POLY);
}

/* gui_gfx ====================================================== */
/*                                                                */
/* ============================================================== */

void gui_gfx(pvr_poly_hdr_t hdr, float u1, float v1, float u2, float v2, int x, int y, float z, int width, int height)
{
  pvr_vertex_t vert; 

  pvr_prim(&hdr, sizeof(pvr_poly_hdr_t)); 
  
  vert.flags = PVR_CMD_VERTEX;
  vert.x = x;
  vert.y = y;
  vert.z = z;
  vert.u = u1;
  vert.v = v1;
  vert.argb = PVR_PACK_COLOR (1, 1, 1, 1);

  vert.oargb = 0;
  pvr_prim(&vert, sizeof(pvr_vertex_t));

  vert.x = x+width;
  vert.u = u2;
  pvr_prim(&vert,sizeof(pvr_vertex_t));

  vert.x = x;
  vert.y = y+height;
  vert.u = u1;
  vert.v = v2;
  pvr_prim(&vert,sizeof(pvr_vertex_t));

  vert.flags = PVR_CMD_VERTEX_EOL;
  vert.x = x+width;
  vert.u = u2;
  pvr_prim(&vert,sizeof(pvr_vertex_t));
}

/* gui_gfx_complete ============================================= */
/*                                                                */
/* ============================================================== */

void gui_gfx_complete(void)
{
  pvr_list_finish();
}

/* gui_text_begin =============================================== */
/*                                                                */
/* ============================================================== */

void gui_text_begin(void)
{
  pvr_list_begin(PVR_LIST_TR_POLY);
}

/* gui_text_complete ============================================ */
/*                                                                */
/* ============================================================== */

void gui_text_complete(void)
{
  pvr_list_finish();
}

/* gui_complete ================================================= */
/*                                                                */
/* ============================================================== */

void gui_complete()
{
  pvr_scene_finish();
}

int cmpdriver(const void *elem1, const void *elem2)
{
  struct GameDriver *driver1 = (struct GameDriver *)elem1;
  struct GameDriver *driver2 = (struct GameDriver *)elem2;

  return(stricmp(driver1->description,driver2->description));
}

/* splash_screen ================================================ */
/*                                                                */
/* ============================================================== */

void splash_screen(wt_font_t *font, uint8 in)
{
  unsigned char *string = "W\4\4ARM\4\2T\4\2OE\2\10TM\n\1\3\3\20\6\20productions";

  int i = 0;

  int   wid,hgt;
  float mult = 1.0;
  float mult_delta = 0.1;

  float fade = 0.0;
  float fade_delta = 0.01;

  if (in)
  {
    mult = 11;
    mult_delta = -0.1;
    fade = 1.0;
    fade_delta = -0.01;
  }

  while(i<100)
  {
    gui_begin();
  
    gui_text_begin();

    font_write(font,string,mult,0,0,0,1,1,1,1,1,&wid,&hgt,0);

    font_write(font,string,mult,(640-wid)/2,(480-hgt)/2,0,1,fade,1,1,1,0,0,1);

    gui_text_complete();

    gui_complete();

    i++;

    mult+=mult_delta;

    fade+=fade_delta;
  }

  return;
}

/* cfg_get_option_max =============================================== */
/*                                                                    */
/* ================================================================== */

void cfg_get_option_max(void)
{
  maim_option_max[0]  = get_config_int("OPTMAX","AUDIO_FREQUENCY",maim_option_max[0]);
  maim_option_max[1]  = get_config_int("OPTMAX","AUDIO_THREAD", maim_option_max[1]);
  maim_option_max[2]  = get_config_int("OPTMAX","AUDIO_THROTTLE", maim_option_max[2]);
  maim_option_max[3]  = get_config_int("OPTMAX","DISPLAY_MODE", maim_option_max[3]);
  maim_option_max[4]  = get_config_int("OPTMAX","SHOW_FPS", maim_option_max[4]);
  maim_option_max[5]  = get_config_int("OPTMAX","CPU_SLICE_SPEEDUP", maim_option_max[5]);
  maim_option_max[6]  = get_config_int("OPTMAX","VIDEO_DMA", maim_option_max[6]);
  maim_option_max[7]  = get_config_int("OPTMAX","VIDEO_THREAD", maim_option_max[7]);
  maim_option_max[8]  = get_config_int("OPTMAX","COLOUR_MODE", maim_option_max[8]);
  maim_option_max[8]  = get_config_int("OPTMAX","COLOR_MODE", maim_option_max[8]);
  maim_option_max[9]  = get_config_int("OPTMAX","VIDEO_THROTTLE", maim_option_max[9]);
  maim_option_max[10] = get_config_int("OPTMAX","FRAMESKIP_LEVEL", maim_option_max[10]);
  maim_option_max[11] = get_config_int("OPTMAX","SCREEN_ORIENTATION", maim_option_max[11]);
  maim_option_max[12] = get_config_int("OPTMAX","CPU0_SPEED", maim_option_max[12]);
  maim_option_max[13] = get_config_int("OPTMAX","CPU1_SPEED", maim_option_max[13]);
  maim_option_max[14] = get_config_int("OPTMAX","CPU2_SPEED", maim_option_max[14]);
}

/* cfg_get_mame_options ============================================= */
/*                                                                    */
/* ================================================================== */

void cfg_get_mame_options(void)
{
  options.color_depth         = get_config_int("MAME","COLOUR_DEPTH",options.color_depth);
  options.color_depth         = get_config_int("MAME","COLOR_DEPTH",options.color_depth);
  options.use_filter          = get_config_int("MAME","USE_FILTER",options.use_filter);
  options.use_samples         = get_config_int("MAME","USE_SAMPLES",options.use_samples);
}

/* main ========================================================= */
/*                                                                */
/* ============================================================== */

int main (int argc, char **argv)
{
  #if defined(NOPS)
  asm("nop");
  #if (NOPS>1)
  asm("nop");
  #if (NOPS>2)
  asm("nop");
  #if (NOPS>3)
  asm("nop");
  #if (NOPS>4)
  asm("nop");
  #if (NOPS>5)
  asm("nop");
  #if (NOPS>6)
  asm("nop");
  #if (NOPS>7)
  asm("nop");
  #if (NOPS>8)
  asm("nop");
  #if (NOPS>9)
  asm("nop");
  #if (NOPS>10)
  asm("nop");
  #if (NOPS>11)
  asm("nop");
  #if (NOPS>12)
  asm("nop");
  #if (NOPS>13)
  asm("nop");
  #if (NOPS>14)
  asm("nop");
  #if (NOPS>15)
  asm("nop");
  #if (NOPS>16)
  asm("nop");
  #if (NOPS>17)
  asm("nop");
  #if (NOPS>18)
  asm("nop");
  #if (NOPS>19)
  asm("nop");
  #if (NOPS>20)
  asm("nop");
  #if (NOPS>21)
  asm("nop");
  #if (NOPS>22)
  asm("nop");
  #if (NOPS>23)
  asm("nop");
  #if (NOPS>24)
  asm("nop");
  #if (NOPS>25)
  asm("nop");
  #if (NOPS>26)
  asm("nop");
  #if (NOPS>27)
  asm("nop");
  #if (NOPS>28)
  asm("nop");
  #if (NOPS>29)
  asm("nop");
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  #endif
  
  unsigned short  found_drivers[MENU_MAX];
  unsigned int    file_sizes[MENU_MAX];
  unsigned char   dc_failures[MENU_MAX];
  double         *prev_fps = 0;

  //int option_follow = 0;
	int option_sel    = 0;
  int option_scroll = 0;
  int option_max    = 0;
  int option_total  = MAIM_OPTIONS_USED;

  int menu_sel      = 0;
  int menu_sel_old  = -1;
  int menu_scroll   = 0;
  int menu_total    = 0;
  int menu_max      = 0;

  int oldbutt=-1;

  int i;
  int repeat = 0;
  int show_snapshot = SNAPSHOT_ENABLED;

  char szStatus[500] = "";

  unsigned char *p = 0;
  unsigned char *just_run = 0;
  int            vid_mode = 0;

  wt_font_t font;

  mallopt(1,0);

  /* ------------ */
  /* Don't ask!!! */
  /* ------------ */

  #if !defined(WT_PROFILE) && defined(CHUNK_BUFFER_SIZE)
  p = (unsigned char *)malloc(CHUNK_BUFFER_SIZE);

  free(p);
  #endif
  
  #if 0
  printf("HERE WE GO...\n");
  {
    uint32 *ptr = 0xFF00001C;
    uint32 val = *ptr;

    printf("0x%08X\n",val);

    //val |= 0x00200000;
    //val |= 0x00000800;
    //val |= 0x00000020;
    val &= ~0xFFFFFFFB;

    *ptr = val;
  }
  printf("...BACK WE COME\n");
  #endif
  
	memset(&options, 0, sizeof(options));

  options.use_samples         = 1;
  options.use_filter          = 1;
	options.color_depth         = 16; /* 16 */
  options.gui_host            = 1;

  options.translucency        = 0;
  options.antialias           = 0;
  options.beam                = 0x00010000;
	
  set_pathlist(FILETYPE_CONFIG,fileio_opts[FILETYPE_CONFIG].deflt);
  set_pathlist(FILETYPE_ROM,fileio_opts[FILETYPE_ROM].deflt);
  set_pathlist(FILETYPE_SAMPLE,fileio_opts[FILETYPE_SAMPLE].deflt);
  set_pathlist(FILETYPE_ARTWORK,fileio_opts[FILETYPE_ARTWORK].deflt);

  #if defined(WT_ROMDISK)
  //set_pathlist(FILETYPE_CONFIG,"/pc/opt/M/cfg");
  set_pathlist(FILETYPE_SCREENSHOT,"/rd/snap");
  #else
  set_pathlist(FILETYPE_SCREENSHOT,fileio_opts[FILETYPE_SCREENSHOT].deflt);
  #endif

  /* ---------------------------- */
  /* Open a configuration file... */
  /* ---------------------------- */

  if (mame_faccess("MDC",FILETYPE_CONFIG))
  {
    set_config_file("MDC");

    ui_mode  = (unsigned char)get_config_int("MDCSTUFF","UI_MODE",ui_mode);
    vid_mode = get_config_int("MDCSTUFF","VID_MODE",vid_mode);
    just_run = strdup(get_config_string("MDCSTUFF","JUST_RUN",just_run));

    cfg_get_option_max();

    cfg_get_mame_options();
  }

  /* ---------------------------- */
  /* Open a more specific file... */
  /* ---------------------------- */

  if (mame_faccess(EXENAME,FILETYPE_CONFIG))
  {
    set_config_file(EXENAME);

    ui_mode  = (unsigned char)get_config_int("MDCSTUFF","UI_MODE",ui_mode);
    vid_mode = get_config_int("MDCSTUFF","VID_MODE",vid_mode);
    just_run = strdup(get_config_string("MDCSTUFF","JUST_RUN",just_run));

    cfg_get_option_max();

    cfg_get_mame_options();
  }

  if (vid_mode) vid_init(vid_mode,PM_RGB565);

  for(i=0;drivers[i];i++)
  {
    int pv = 0;

    if (menu_total < MENU_MAX && ((pv=mame_faccess(drivers[i]->name,FILETYPE_ROM))))
    {
      char name[256];

      FILE *fp;

      sprintf(name, "%s/%s.zip", get_path_for_filetype(FILETYPE_ROM,pv-1,0), drivers[i]->name);

      if ((fp = fopen(name,"rb")))
      {
        fseek( fp, 0, SEEK_END );
        file_sizes[menu_total] = ftell( fp );
        fclose(fp);
      }

      dc_failures[menu_total] = 0;

      found_drivers[menu_total++] = i;

      if (just_run && !stricmp(drivers[i]->name,just_run))
        menu_sel = menu_total;

      printf("F  %s (%s)\n",drivers[i]->name,name);
    }
    else printf("NF %s\n",drivers[i]->name);
  }

  free(just_run);

  if (!menu_sel)
    just_run = 0;

  if ((prev_fps = malloc(menu_total*sizeof(double))))
    memset(prev_fps,0x00,menu_total*sizeof(double));

  osd_init();

  if (gui_init(&font,1))
  {
    menu_max = (MENU_HEIGHT/(font.height*MENU_FONT));
    option_max = (OPTIONS_HEIGHT/(font.height*OPTIONS_FONT));
  }

  while(menu_total)
  {
    if ((first_cont_state = (cont_state_t *)maple_dev_status(first_controller)))
    {
      struct GameDriver *curdriver = (struct GameDriver *)drivers[found_drivers[menu_sel]];

      if (!just_run)
      {
        gui_begin();

        gui_gfx_begin();
         
        /* ------------------- */
        /* Backdrop processing */
        /* ------------------- */

        if (txr_back)
          gui_gfx(hdr_back,0,0,0.625,0.9375,BACKDROP_X,BACKDROP_Y,BACKDROP_Z,BACKDROP_WIDTH,BACKDROP_HEIGHT);

        /* ------------------- */
        /* Snapshot processing */
        /* ------------------- */

        if (txr_snap && menu_sel != menu_sel_old && first_cont_state->buttons == 0)
        {
          if (show_snapshot != -1)
          {
            struct GameDriver *usedriver = curdriver;

            char name[256];

            show_snapshot = 0;

            /* -------------------------------------- */
            /* Loop through clones to find a snapshot */
            /* -------------------------------------- */

            while(usedriver && !show_snapshot)
            {
              int pv = 0;

              if ((pv=mame_faccess(usedriver->name,FILETYPE_SCREENSHOT)))
              {
                sprintf(name,"%s/%s.png",get_path_for_filetype(FILETYPE_SCREENSHOT,pv-1,0),usedriver->name);
                show_snapshot = 1;
              }
              else
                usedriver = (struct GameDriver *)usedriver->clone_of;
            }

            if (show_snapshot)
            {
              printf("%s %d\n",__FILE__,__LINE__);

              png_to_texture(name,txr_snap,PNG_NO_ALPHA);

              printf("%s %d\n",__FILE__,__LINE__);
            }
            else
            {
              memset(txr_snap,0x00,256*256*2);

              /* ----------------- */
              /* No snapshot found */
              /* ----------------- */
            }

          }

          menu_sel_old = menu_sel;
        }

        /* ----------------------------------------- */
        /* If we have a snapshot to show, show it... */
        /* ----------------------------------------- */

        if (show_snapshot == 1)
        {
          gui_gfx(hdr_snap,0,0,1,1,SNAPSHOT_X,SNAPSHOT_Y,SNAPSHOT_Z,SNAPSHOT_WIDTH,SNAPSHOT_HEIGHT);
        }

        gui_gfx_complete();
        
        gui_text_begin();

        /* -------------------- */
        /* Game Menu processing */
        /* -------------------- */

        for(i=0;i+menu_scroll < menu_total && i < menu_max ;i++)
        {
          font_write(&font,(char *)drivers[found_drivers[i+menu_scroll]]->description,MENU_FONT,
                     MENU_X,MENU_Y+(i*(font.height*MENU_FONT)),MENU_X+MENU_WIDTH,MENU_Z,
                     ((i+menu_scroll)==menu_sel)?1:0.5f,MENU_RED,MENU_GREEN,MENU_BLUE,
                     0,0,1);
        }

        /* ----------------- */
        /* Option processing */
        /* ----------------- */

        for(i=0;i+option_scroll < option_total && i < option_max ;i++)
        {
          sprintf(szStatus,"%s %d",maim_option_name[i+option_scroll],maim_option_cur[i+option_scroll]);
          font_write(&font,szStatus,OPTIONS_FONT,
                     OPTIONS_X,OPTIONS_Y+(i*(font.height*OPTIONS_FONT)),OPTIONS_X+OPTIONS_WIDTH,OPTIONS_Z,
                     ((i+option_scroll)==option_sel)?1:0.5f,OPTIONS_RED,OPTIONS_GREEN,OPTIONS_BLUE,
                     0,0,1);
        }

        /* -------------------- */
        /* Game info processing */
        /* -------------------- */

        i = 2;

        sprintf(szStatus, "%s\n%s", curdriver->manufacturer, curdriver->year);
        font_write(&font,szStatus,INFO_FONT,
                   INFO_X,INFO_Y,INFO_X+INFO_WIDTH,INFO_Z,
                   0.5, INFO_RED, INFO_GREEN, INFO_BLUE,
                   0,0,1);

        if (curdriver->clone_of && curdriver->clone_of->description)
        {
          sprintf(szStatus, "Clone of %s", curdriver->clone_of->description);
          font_write(&font, szStatus, INFO_FONT,
                     INFO_X, INFO_Y+(i++*(font.height*INFO_FONT)), INFO_X+INFO_WIDTH, INFO_Z,
                     0.5, INFO_RED, INFO_GREEN, INFO_BLUE,
                     0,0,1);
        }

        if (curdriver->flags & GAME_NOT_WORKING)
        {
          font_write(&font, "Game not working", INFO_FONT,
                     INFO_X, INFO_Y+(i++*(font.height*INFO_FONT)), INFO_X+INFO_WIDTH, INFO_Z,
                     0.5, INFO_RED, INFO_GREEN, INFO_BLUE,
                     0,0,1);
        }

        i++;

        sprintf(szStatus, "Zip size %d bytes", file_sizes[menu_sel]);
        font_write(&font, szStatus, INFO_FONT,
                   INFO_X, INFO_Y+(i++*(font.height*INFO_FONT)), INFO_X+INFO_WIDTH, INFO_Z,
                   (file_sizes[menu_sel]>7000000)?1:0.5, INFO_RED, INFO_GREEN, INFO_BLUE,
                   0,0,1);

        /* --------------------------- */
        /* Performance info processing */
        /* --------------------------- */
          
        if (prev_fps && prev_fps[menu_sel])
        {
          sprintf(szStatus, "Previous Speed %.2f FPS",prev_fps[menu_sel]);
          font_write(&font, szStatus, INFO_FONT,
                     INFO_X, INFO_Y+(i++*(font.height*INFO_FONT)), INFO_X+INFO_WIDTH, INFO_Z,
                     0.5, INFO_RED, INFO_GREEN, INFO_BLUE,
                     0,0,1);
        }

        /* ------------------ */
        /* Failure processing */
        /* ------------------ */
          
        szStatus[0] = 0;

        switch(dc_failures[menu_sel])
        {
          case 1:
            sprintf(szStatus,"Bad ROMS in %s.zip",curdriver->name);
            break;

          case 2:
            sprintf(szStatus,"Memory init failed");
            break;

          case 3:
            sprintf(szStatus,"osd_init failed");
            break;

          case 4:
            sprintf(szStatus,"Video emulation failed");
            break;
        }

        if (szStatus[0])
        {
          i++;

          font_write(&font, szStatus, INFO_FONT,
                     INFO_X, INFO_Y+(i++*(font.height*INFO_FONT)), INFO_X+INFO_WIDTH, INFO_Z,
                     0.5, INFO_FAIL_RED, INFO_FAIL_GREEN, INFO_FAIL_BLUE,
                     0,0,1);
        }

        /* --------------------------------------------------------- */
        /* Special messages - only applicable for development builds */
        /* --------------------------------------------------------- */

        #if defined(WT_ASSERT)
        sprintf(szStatus,"*** ASSERTIONS IN PLACE - REMOVE THEM!!!! ***");
        font_write(&font,szStatus,0.75,50,50,0,2.5,1,1,1,1,0,0,1);
        #endif

        gui_text_complete();

        gui_complete();

        /* --------------------- */
        /* User input processing */
        /* --------------------- */

        if (oldbutt && first_cont_state->buttons == oldbutt)
          repeat++;
        else
          repeat=0;
      }
      else // just_run processing
      {
        /* ------------------------------------------------- */
        /* If we return from running the game, we want a GUI */
        /* ------------------------------------------------- */

        just_run = 0;

        /* ----------------------------------------------------- */
        /* Actual menu selection is 1 less than the value stored */
        /* ----------------------------------------------------- */

        menu_sel--;

        /* ----------------------------------------------- */
        /* Fake a 'default options' and 'start' keypresses */
        /* ----------------------------------------------- */

        first_cont_state->buttons = CONT_X | CONT_START;

        oldbutt = 0;
      }

      if (repeat>50 || first_cont_state->buttons!=oldbutt)
      {
        if (first_cont_state->buttons & CONT_B)
        {
          #if 1
          menu_sel      = 0;
          menu_sel_old  = -1;
          menu_scroll   = 0;
          menu_total    = 0;

          for(i=0;drivers[i];i++)
          {
            int pv = 0;

            if (menu_total < MENU_MAX && ((pv=mame_faccess(drivers[i]->name,FILETYPE_ROM))))
            {
              char name[256];

              FILE *fp;

              sprintf(name, "%s/%s.zip", get_path_for_filetype(FILETYPE_ROM,pv-1,0), drivers[i]->name);

              if ((fp = fopen(name,"rb")))
              {
                fseek( fp, 0, SEEK_END );
                file_sizes[menu_total] = ftell( fp );
                fclose(fp);
              }

              dc_failures[menu_total] = 0;

              found_drivers[menu_total++] = i;

              printf("F  %s (%s)\n",drivers[i]->name,name);
            }
            else printf("NF %s\n",drivers[i]->name);
          }
          #else
          {
            FILE *fp = 0;

            sprintf(szStatus,"# %s_TINY_OPTIONS={%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d}\n",
                    curdriver->name,
                    maim_option_cur[0],maim_option_cur[1],maim_option_cur[2],maim_option_cur[3],maim_option_cur[4],
                    maim_option_cur[5],maim_option_cur[6],maim_option_cur[7],maim_option_cur[8],maim_option_cur[9],
                    maim_option_cur[10],maim_option_cur[11]);
            
            fp = fopen("/pc/opt/M/TO.txt","wa");

            if (fp)
            {
              fprintf(fp,szStatus);
              fclose(fp);
            }
            else
              printf(szStatus);
          }
          #endif
        }
        else
        {
          if(first_cont_state->buttons & CONT_DPAD_UP)  menu_sel--;
          if(first_cont_state->buttons & CONT_DPAD_DOWN) menu_sel++;

          if(menu_sel<0) menu_sel=0;
          if(menu_sel>=menu_total) menu_sel=menu_total-1;
          
          if (menu_sel>(menu_scroll+menu_max-1)) menu_scroll++;
          if (menu_sel<menu_scroll) menu_scroll--;

          if(first_cont_state->buttons & CONT_X)
          {
            /* ------------------------------------------------ */
            /* Read in override default options from a cfg file */
            /* ------------------------------------------------ */

            if (mame_faccess(drivers[found_drivers[menu_sel]]->name,FILETYPE_CONFIG))
            {
              int opt;

              set_config_file((char *)drivers[found_drivers[menu_sel]]->name);

              maim_option_cur[0]  = get_config_int("DEFAULTS","AUDIO_FREQUENCY",maim_option_def[0]);
              maim_option_cur[1]  = get_config_int("DEFAULTS","AUDIO_THREAD", maim_option_def[1]);
              maim_option_cur[2]  = get_config_int("DEFAULTS","AUDIO_THROTTLE", maim_option_def[2]);
              maim_option_cur[3]  = get_config_int("DEFAULTS","DISPLAY_MODE", maim_option_def[3]);
              maim_option_cur[4]  = get_config_int("DEFAULTS","SHOW_FPS", maim_option_def[4]);
              maim_option_cur[5]  = get_config_int("DEFAULTS","CPU_SLICE_SPEEDUP", maim_option_def[5]);
              maim_option_cur[6]  = get_config_int("DEFAULTS","VIDEO_DMA", maim_option_def[6]);
              maim_option_cur[7]  = get_config_int("DEFAULTS","VIDEO_THREAD", maim_option_def[7]);
              maim_option_cur[8]  = get_config_int("DEFAULTS","COLOUR_MODE", maim_option_def[8]);
              maim_option_cur[8]  = get_config_int("DEFAULTS","COLOR_MODE", maim_option_cur[8]);
              maim_option_cur[9]  = get_config_int("DEFAULTS","VIDEO_THROTTLE", maim_option_def[9]);
              maim_option_cur[10] = get_config_int("DEFAULTS","FRAMESKIP_LEVEL", maim_option_def[10]);
              maim_option_cur[11] = get_config_int("DEFAULTS","SCREEN_ORIENTATION",maim_option_def[11]);
              maim_option_cur[12] = get_config_int("DEFAULTS","CPU0_SPEED", maim_option_def[12]);
              maim_option_cur[13] = get_config_int("DEFAULTS","CPU1_SPEED", maim_option_def[13]);
              maim_option_cur[14] = get_config_int("DEFAULTS","CPU2_SPEED", maim_option_def[14]);

              for(opt=0;opt<MAIM_OPTIONS;opt++)
                if (maim_option_cur[opt] < 0 || maim_option_cur[opt] > maim_option_max[opt])
                  maim_option_cur[opt] = maim_option_def[opt];

              cfg_get_mame_options();
            }
            else
              memcpy(&maim_option_cur,&maim_option_def,MAIM_OPTIONS*sizeof(int));
          }

          if(first_cont_state->buttons & CONT_DPAD_RIGHT)
            maim_option_cur[option_sel] = (maim_option_cur[option_sel]+maim_option_inc[option_sel]);
          if(first_cont_state->buttons & CONT_DPAD_LEFT)
            maim_option_cur[option_sel] = (maim_option_cur[option_sel]-maim_option_inc[option_sel]);

          if (maim_option_cur[option_sel]>maim_option_max[option_sel])
            maim_option_cur[option_sel] = 0;

          if (maim_option_cur[option_sel]<0)
            maim_option_cur[option_sel] = maim_option_max[option_sel];
          
          if (first_cont_state->buttons & CONT_Y) option_sel--;
          if (first_cont_state->buttons & CONT_A) option_sel++;
      

          if (option_sel<0) option_sel=0;
          if (option_sel>=option_total) option_sel=option_total-1;

          if (option_sel>(option_scroll+option_max-1)) option_scroll++;
          if (option_sel<option_scroll) option_scroll--;

          //if (option_follow && menu_sel != menu_sel_old)
          //  memcpy(&maim_option,&maim_option_def[found_drivers[menu_sel]],MAIM_OPTIONS*sizeof(int));

          if(first_cont_state->buttons & CONT_START)
          {
            use_throttle       = maim_option_cur[9];

            use_audio_throttle = maim_option_cur[2];

            options.samplerate = maim_option_cur[0];

            if (options.samplerate)
            {
              play_sound = 1;
              use_audio_thread = maim_option_cur[1];
            }
            else
            {
              play_sound = 0;
              use_audio_thread = 1; /* 1 = NO! */

              /* if (use_audio_throttle) use_throttle = 1; */
            }

            use_display_mode              = maim_option_cur[3];
            showfps                       = maim_option_cur[4];
            cpu_slices_per_frame_override = maim_option_cur[5];
            use_dma                       = maim_option_cur[6];
            use_video_thread              = maim_option_cur[7];
            screen_orientation            = maim_option_cur[11];

            /* -------------------------------- */
            /* force static colours in DMA mode */
            /* -------------------------------- */

            if (use_dma)
              global_colormode = 3;
            else
              global_colormode = maim_option_cur[8]-1;

            global_frameskip = frameskip = maim_option_cur[10];

            cpu_speed[0] = (double)maim_option_cur[12]/100.0;
            cpu_speed[1] = (double)maim_option_cur[13]/100.0;
            cpu_speed[2] = (double)maim_option_cur[14]/100.0;

            autoframeskip = 0;

            if (frameskip == 12)
            {
              frameskip = 0;
              autoframeskip = 1;
            }

            #ifdef WT_PROFILE
            memset(&totalticks,0,100*sizeof(unsigned long));
            memset(&totalinstances,0,100*sizeof(unsigned long));
            cycles=0;
            #endif

            gui_shutdown(&font,0);

            global_dc_failure = 0;

            #if 0
            {
              struct MachineDriver *drv = drivers[found_drivers[menu_sel]]->drv;

              //drv->frames_per_second = maim_option_cur[13];
              drv->vblank_duration = maim_option_cur[14];
            }
            #endif
            
            run_game(found_drivers[menu_sel]);

            dc_failures[menu_sel] = global_dc_failure;

            if (prev_fps)
              prev_fps[menu_sel] = (double)avg_fps;

            menu_sel_old=-1; /* Force a reload of the snap */

            printf("JS_CALLS: %lu\n",js_calls);

            printf("DEBUG_VALUE: %lu\n",dbgvalue);

            dbgvalue = 0;

            //thd_pslist();

            gui_init(&font,0);
          }
        }

        if (repeat<=50)
          oldbutt = first_cont_state->buttons;
      }

  	  if ((first_cont_state->ltrig > 16) && (first_cont_state->rtrig > 16))
      {
        gui_shutdown(&font,1);
  	    break;
      }
    }
  }

  /* ---------------- */
  /* No roms found... */
  /* ---------------- */

  while(!menu_total)
  {
    if ((first_cont_state = (cont_state_t *)maple_dev_status(first_controller)))
    {
      int x = NOROM_X;
      int y = NOROM_Y;

      gui_begin();

      gui_gfx_begin();

      /* Any graphics could go in here */

      gui_gfx_complete();
      
      gui_text_begin();

      sprintf(szStatus,"No ROMS found in /cd/maim or /cd/roms");

      font_write(&font, szStatus, NOROM_FONT, x, y, 0, 
                 NOROM_Z, 1, NOROM_RED, NOROM_GREEN, NOROM_BLUE, 0,0,1);

      gui_text_complete();

      gui_complete();

  	  //if ((first_cont_state->ltrig > 16) && (first_cont_state->rtrig > 16))
      {
        gui_shutdown(&font,1);
  	    break;
      }
    }
    else
    {
      printf("No ROMs & No Controllers... TIME TO QUIT\n");

      gui_shutdown(&font,1);
      break;
    }
  }

  free(prev_fps);

	osd_exit();
	
  printf("md_calls: %lu\n",md_calls);
  printf("KP_COUNT: %lu\n",kp_count);

	exit(0);
}

void CLIB_DECL logerror(const char *text,...)
{
  #if 0
	FILE *log;
	log = fopen("/pc/tmp/mame.log", "a");
	if(log) 
	{
		fprintf(log, "%s\r\n", text);
		fclose(log);
	}
  #endif
}
