#include "osdepend.h"
#include "driver.h"
#include "unzip.h"
#ifndef MESS
  #include <signal.h>

/* Verbose outputs to error.log ? */
  #define VERBOSE 	1

/* Use the file cache ? */
  #define FILE_CACHE	0

  #if VERBOSE
    #define LOG(x)	printf x
  #else
    #define LOG(x)	/* x */
  #endif

char *roms = NULL;
char **rompathv = NULL;
int rompathc = 0;

char *samples = NULL;
char **samplepathv = NULL;
int samplepathc = 0;

char *cfgdir = "/cd/cfg", *nvdir = "/cd/nvram", *hidir = "/cd/hi", *inpdir, *stadir;
char *memcarddir, *artworkdir = "/cd/artwork", *screenshotdir = "/cd/snap";
char *cheatdir; /* Steph */

char *alternate_name;          /* for "-romdir" */

typedef enum
{
  kPlainFile,
  kRAMFile,
  kZippedFile
} eFileType;

typedef struct
{
  FILE *file;
  unsigned char *data;
  unsigned int offset;
  unsigned int length;
  eFileType type;
  unsigned int crc;
} FakeFileHandle;


extern unsigned int crc32 (unsigned int crc, const unsigned char *buf, unsigned int len);
static int checksum_file (const char *file, unsigned char **p, unsigned int *size, unsigned int *crc);

/* stat ========================================================= */
/*                                                                */
/* ============================================================== */

int stat(char *path, struct mame_stat *fileinfo)
{
  file_t fd = fs_open(path,O_RDONLY);

  if (fd)
  {
    fileinfo->st_size = fs_total(fd);

    fs_close(fd);
  }

  return(!fd);
}

/* stat_dir ===================================================== */
/*                                                                */
/* ============================================================== */

int stat_dir(char *path, struct mame_stat *fileinfo)
{
  file_t fd = fs_open(path,O_DIR);

  if (fd)
  {
    fileinfo->st_mode = S_IFDIR ;

    fs_close(fd);
  }

  return(!fd);
}

/* raise ======================================================== */
/*                                                                */
/* ============================================================== */

int raise(int signal)
{
  return 0; /* SUCCESS */
}

/* decompose_rom_sample_path ==================================== */
/*                                                                */
/* This function can be called several times with different       */
/* parameters, for example by "mame -verifyroms *".               */
/*                                                                */
/* ============================================================== */

void decompose_rom_sample_path (char *rompath, char *samplepath)
{
  char *token;

  /* ------------------------------- */
  /* Start with zero path components */
  /* ------------------------------- */

  rompathc = samplepathc = 0;

  /* --------------------------- */
  /* realloc with NULL == malloc */
  /* --------------------------- */

  roms    = realloc(roms,strlen(rompath)+1);
  samples = realloc(samples,strlen(samplepath)+1);

  /* ---------------------------------------- */
  /* Problems with the allocation? return now */
  /* ---------------------------------------- */

  if (!roms || !samples)
  {
    free(roms);
    free(samples);
    logerror("decompose_rom_sample_path: failed to malloc!\n");
    raise(SIGABRT);
    return;
  }

  strcpy (roms, rompath);

  /* ----------------------------------------------- */
  /* Parse the rompath, adding entries along the way */
  /* ----------------------------------------------- */

  token = strtok(roms,";");

  while (token)
  {
    if (!((rompathv = realloc(rompathv,(rompathc+1)*sizeof(char*)))))
      break;

    rompathv[rompathc++] = token;

    token = strtok(NULL,";");
  }

  strcpy(samples, samplepath);

  /* -------------------------------------------------- */
  /* Parse the samplepath, adding entries along the way */
  /* -------------------------------------------------- */

  token = strtok(samples,";");

  while (token)
  {
    if (!((samplepathv = realloc(samplepathv,(samplepathc+1)*sizeof(char*)))))
      break;

    samplepathv[samplepathc++] = token;

    token = strtok(NULL,";");
  }
}

/*
 * file handling routines
 *
 * gamename holds the driver name, filename is only used for ROMs and samples.
 * if 'write' is not 0, the file is opened for write. Otherwise it is opened
 * for read.
 */

/* osd_faccess ================================================== */
/*                                                                */
/* Check if roms/samples for a game exist at all return index+1   */
/* of the path vector component on success, otherwise 0.          */
/*                                                                */
/* ============================================================== */

int osd_faccess (const char *newfilename, int filetype)
{
  static int indx;
  static const char *filename;
  char name[256];
  char **pathv;
  int pathc;
  char *dir_name;

  /* ---------------------------------------- */
  /* if filename == NULL, continue the search */
  /* ---------------------------------------- */

  if (newfilename)
  {
    indx = 0;
    filename = newfilename;
  }
  else
    indx++;

#ifdef MESS
  if (filetype == FILETYPE_ROM || filetype == FILETYPE_IMAGE_R || filetype == FILETYPE_IMAGE_RW)
#else
  if (filetype == FILETYPE_ROM)
#endif
  {
    pathv = rompathv;
    pathc = rompathc;
  }
  else if (filetype == FILETYPE_SAMPLE)
  {
    pathv = samplepathv;
    pathc = samplepathc;
  }
  else if (filetype == FILETYPE_SCREENSHOT || filetype == FILETYPE_CONFIG)
  {
    void *f = 0;

    switch(filetype)
    {
      case FILETYPE_SCREENSHOT:
        sprintf(name, "%s/%s.png", screenshotdir, newfilename);
        break;

      case FILETYPE_CONFIG:
        sprintf(name, "%s/%s.cfg", cfgdir, newfilename);
        break;
    }

    //printf("FILE: %s\n",name);

    if ((f = fopen (name, "rb")))
    {
      fclose (f);
      return 1;
    }
    else
      return 0;
  }
  else
    return 0;

  for (;indx<pathc;indx++)
  {
    struct mame_stat stat_buffer;

    dir_name = pathv[indx];

    /* -------------------------------------- */
    /* Does such a directory (or file) exist? */
    /* -------------------------------------- */

    sprintf (name, "%s/%s", dir_name, filename);

    if (!stat_dir(name, &stat_buffer))
      return(indx+1);

#ifndef WT_NOUNZIP

    /* ------------------------------- */
    /* Try again with a .zip extension */
    /* ------------------------------- */

    sprintf (name, "%s/%s.zip", dir_name, filename);

    if (!stat(name,&stat_buffer))
      return(indx+1);

    /* ------------------------------- */
    /* Try again with a .zif extension */
    /* ------------------------------- */

    sprintf (name, "%s/%s.zif", dir_name, filename);

    if (!stat(name, &stat_buffer))
      return(indx+1);

#endif /* WT_NOUNZIP */
  }

  /* -------- */
  /* No match */
  /* -------- */

  return 0;
}

/* osd_fopen ==================================================== */
/*                                                                */
/* ============================================================== */

void *osd_fopen (const char *game, const char *filename, int filetype, int _write)
{
  char name[256];
  char *gamename;
  int found = 0;
  int indx;
  struct mame_stat stat_buffer;
  FakeFileHandle *f;
  int pathc = 0;
  char **pathv = 0;

  /* -------------------------------- */
  /* Allocate a file handle structure */
  /* -------------------------------- */

  if (!((f = (FakeFileHandle *) malloc (sizeof (FakeFileHandle)))))
  {
    logerror("osd_fopen: failed to malloc FakeFileHandle!\n");
    return 0;
  }

  /* ------------- */
  /* Clear the FFH */
  /* ------------- */

  memset(f,0x00,sizeof(FakeFileHandle));

  gamename = (char *)game;

  /* ----------------------- */
  /* Support "-romdir" yuck. */
  /* ----------------------- */

  if (alternate_name)
  {
    LOG(("osd_fopen: -romdir overrides '%s' by '%s'\n", gamename, alternate_name));
    gamename = alternate_name;
  }

  /* -------------------------------------- */
  /* What type of file are we dealing with? */
  /* -------------------------------------- */

  switch (filetype)
  {
    case FILETYPE_ROM:
      if (!pathv)
      {
        LOG(("osd_fopen: using rompath\n"));

        pathc = rompathc;
        pathv = rompathv;
      }
    case FILETYPE_SAMPLE:
      if (!pathv)
      {
        LOG(("osd_fopen: using samplepath\n"));

        pathc = samplepathc;
        pathv = samplepathv;
      }

      if (_write)
      {
        logerror("osd_fopen: type %02x write not supported\n",filetype);
        break;
      }

      for (indx=0;indx<pathc && !found;++indx)
      {
        const char *dir_name = pathv[indx];

        if (!found)
        {
          sprintf(name,"%s/%s",dir_name,gamename);

          LOG(("Trying directory %s\n", name));

          /* ------------------------ */
          /* Try for a directory name */
          /* ------------------------ */

          if (!stat_dir(name, &stat_buffer) && (stat_buffer.st_mode & S_IFDIR))
          {
            /* --------------------- */
            /* OK - Now the filename */
            /* --------------------- */

            sprintf (name,"%s/%s/%s",dir_name,gamename,filename);

            LOG(("Trying %s\n", name));

            if (filetype != FILETYPE_ROM || !checksum_file(name,0,&f->length,&f->crc))
            {
              f->type = kPlainFile;
              f->file = fopen(name,"rb");
              found = f->file != 0;
            }
          }
        }

        #ifndef WT_NOUNZIP

        if (!found)
        {
          const char *dir_name = pathv[indx];

          /* ------------------------- */
          /* Try with a .zip extension */
          /* ------------------------- */

          sprintf(name,"%s/%s.zip",dir_name,gamename);

          LOG(("Trying %s file\n", name));

          if (!stat(name,&stat_buffer))
          {
            if (!load_zipped_file(name,filename,&f->data,&f->length))
            {
              LOG(("Using (osd_fopen) zip file for %s\n", filename));

              f->type = kZippedFile;
              f->offset = 0;
              f->crc = crc32(0L, f->data, f->length);
              found = 1;
            }
          }
        }

        if (!found)
        {
          const char *dir_name = pathv[indx];

          /* ---------------------------------------------------- */
          /* Try with a .zip directory (if ZipMagic is installed) */
          /* ---------------------------------------------------- */

          sprintf (name, "%s/%s.zip", dir_name, gamename);

          LOG(("Trying %s directory\n", name));

          if (!stat_dir(name,&stat_buffer) && (stat_buffer.st_mode & S_IFDIR))
          {
            sprintf (name, "%s/%s.zip/%s", dir_name, gamename, filename);

            if (filetype == FILETYPE_ROM)
            {
              if (!checksum_file(name, &f->data, &f->length, &f->crc) == 0)
              {
                f->type = kRAMFile;
                f->offset = 0;
                found = 1;
              }
            }
            else
            {
              f->type = kPlainFile;
              f->file = fopen (name, "rb");
              found = f->file != 0;
            }
          }
        }
        #endif /* WT_NOUNZIP */
      }
      break;

#ifdef MESS

    case FILETYPE_IMAGE_R:

      if (_write)
      {
        logerror("osd_fopen: type %02x write not supported\n",filetype);
        break;
      }

      LOG(("osd_fopen: using rompath\n"));

      pathc = rompathc;
      pathv = rompathv;

      LOG(("Open IMAGE_R '%s' for %s\n", filename, game));

      for (indx = 0; indx < pathc && !found; ++indx)
      {
        const char *dir_name = pathv[indx];

        /* ---------------------------------------- */
        /* This section allows exact path from .cfg */
        /* ---------------------------------------- */

        if (!found)
        {
          sprintf(name,"%s",dir_name);

          if (!stat_dir(name,&stat_buffer) && (stat_buffer.st_mode & S_IFDIR))
          {
            sprintf(name,"%s/%s",dir_name,filename);

            f->type = kPlainFile;
            f->file = fopen(name,"rb");
            found = f->file!=0;
          }
        }

        if (!found)
        {
          sprintf (name, "%s/%s", dir_name, gamename);

          LOG(("Trying %s directory\n", name));

          if (!stat_dir(name, &stat_buffer) && (stat_buffer.st_mode & S_IFDIR))
          {
            sprintf (name, "%s/%s/%s", dir_name, gamename, filename);

            LOG(("Trying %s file\n", name));

            f->type = kPlainFile;
            f->file = fopen (name, "rb");
            found = f->file != 0;
          }
        }

        #ifndef WT_NOUNZIP

        /* ------------------------- */
        /* Zip cart support for MESS */
        /* ------------------------- */

        if (!found)
        {
          char *extension = strrchr (name, '.');    /* find extension */

          if (extension)
            strcpy (extension, ".zip");
          else
            strcat (name, ".zip");

          LOG(("Trying %s file\n", name));

          if (!stat(name, &stat_buffer))
          {
            if (!load_zipped_file(name, filename, &f->data, &f->length))
            {
              LOG(("Using (osd_fopen) zip file for %s\n", filename));

              f->type = kZippedFile;
              f->offset = 0;
              f->crc = crc32(0L, f->data, f->length);
              found = 1;
            }
          }
        }

        if (!found)
        {
          /* ------------------------- */
          /* Try with a .zip extension */
          /* ------------------------- */

          sprintf(name,"%s/%s.zip",dir_name,gamename);

          LOG(("Trying %s file\n", name));

          if (!stat(name, &stat_buffer))
          {
            if (!load_zipped_file(name, filename, &f->data, &f->length))
            {
              LOG(("Using (osd_fopen) zip file for %s\n", filename));

              f->type = kZippedFile;
              f->offset = 0;
              f->crc = crc32(0L, f->data, f->length);
              found = 1;
            }
          }
        }

        #endif /* WT_NOUNZIP */
      }
      break; /* end of IMAGE_R */

    case FILETYPE_IMAGE_RW:
    {
      static char *write_modes[] = {"rb","wb","r+b","r+b","w+b"};

      char file[256];
      char *extension;

      LOG(("Open IMAGE_RW '%s' for %s mode '%s'\n", filename, game, write_modes[_write]));

      strcpy (file, filename);

      do
      {
        /* --------------------------------------------- */
        /* 29-05-00 Lee Ward: Reversed the search order. */
        /* to prevent new files being created in the     */
        /* application root by default.                  */
        /* --------------------------------------------- */

        for (indx=rompathc-1; indx>=0 && !found; --indx)
        {
          const char *dir_name = rompathv[indx];

          /* ------------------ */
          /* Exact path support */
          /* ------------------ */

          if (!found)
          {
            sprintf (name, "%s/%s", dir_name, gamename);

            LOG(("Trying %s directory\n", name));

            if (!stat_dir(name, &stat_buffer) && (stat_buffer.st_mode & S_IFDIR))
            {
              sprintf (name, "%s/%s/%s", dir_name, gamename, file);

              LOG(("Trying %s file\n", name));

              f->file = fopen (name, write_modes[_write]);

              found = f->file != 0;

              if (!found && _write == 3)
              {
                f->file = fopen(name, write_modes[4]);
                found = f->file != 0;
              }
            }
          }

          #ifndef WT_NOUNZIP

          /* ---------------------------------------- */
          /* Steph - Zip disk images support for MESS */
          /* ---------------------------------------- */

          if (!found && !_write)
          {
            extension = strrchr (name, '.');
            
            /* -------------------- */
            /* add .zip for zipfile */
            /* -------------------- */

            if (extension)
              strcpy(extension, ".zip");
            else
              strcat(extension, ".zip");

            LOG(("Trying %s file\n", name));

            if (!stat(name, &stat_buffer))
            {
              if (load_zipped_file(name, filename, &f->data, &f->length) == 0)
              {
                LOG(("Using (osd_fopen) zip file for %s\n", filename));

                f->type = kZippedFile;
                f->offset = 0;
                f->crc = crc32(0L, f->data, f->length);
                found = 1;
              }
            }
          }

          #endif /* WT_NOUNZIP */

          if (!found)
          {
            sprintf(name, "%s", dir_name);

            LOG(("Trying %s directory\n", name));

            if (!stat_dir(name,&stat_buffer) && (stat_buffer.st_mode & S_IFDIR))
            {
              sprintf(name,"%s/%s", dir_name, file);

              LOG(("Trying %s file\n", name));

              f->file = fopen(name, write_modes[_write]);
              found = f->file != 0;

              if (!found && _write == 3)
              {
                f->file = fopen(name, write_modes[4]);
                found = f->file != 0;
              }
            }
          }

          #ifndef WT_NOUNZIP

          if (!found && !_write)
          {
            extension = strrchr (name, '.');

            /* -------------------- */
            /* add .zip for zipfile */
            /* -------------------- */

            if (extension)
              strcpy(extension, ".zip");
            else
              strcat(extension, ".zip");

            LOG(("Trying %s file\n", name));

            if (!stat(name, &stat_buffer))
            {
              if (!load_zipped_file(name, filename, &f->data, &f->length))
              {
                LOG(("Using (osd_fopen) zip file for %s\n", filename));

                f->type = kZippedFile;
                f->offset = 0;
                f->crc = crc32(0L, f->data, f->length);
                found = 1;
              }
            }
          }

          if (!found && !_write)
          {
            sprintf (name, "%s/%s.zip", dir_name, gamename);

            LOG(("Trying %s file\n", name));

            if (!stat(name, &stat_buffer))
            {
              if (!load_zipped_file(name, file, &f->data, &f->length))
              {
                LOG(("Using (osd_fopen) zip file for %s\n", filename));

                f->type = kZippedFile;
                f->offset = 0;
                f->crc = crc32 (0L, f->data, f->length);
                found = 1;
              }
            }
          }

          if (!found)
          {
            /* try with a .zip directory (if ZipMagic is installed) */
            
            sprintf (name, "%s/%s.zip", dir_name, gamename);
            
            LOG(("Trying %s ZipMagic directory\n", name));
            
            if (!stat_dir(name, &stat_buffer) && (stat_buffer.st_mode & S_IFDIR))
            {
              sprintf (name, "%s/%s.zip/%s", dir_name, gamename, file);

              LOG(("Trying %s\n", name));

              f->file = fopen (name, write_modes[_write]);
              found = f->file != 0;

              if (!found && _write == 3)
              {
                f->file = fopen(name, write_modes[4]);
                found = f->file != 0;
              }
            }
          }

          #endif /* WT_NOUNZIP */

          if (found)
            LOG(("IMAGE_RW %s FOUND in %s!\n", file, name));
        }

        extension = strrchr (file, '.');

        if (extension)
          *extension = '\0';

      } while (!found && extension);
    }
    break;
    #endif /* MESS */


    case FILETYPE_NVRAM:
      if (!found)
      {
        sprintf (name, "%s/%s.nv", nvdir, gamename);

        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      #ifndef WT_NOUNZIP
      
      if (!found)
      {
        /* ---------------------------------------------------- */
        /* try with a .zip directory (if ZipMagic is installed) */
        /* ---------------------------------------------------- */

        sprintf (name, "%s.zip/%s.nv", nvdir, gamename);
        
        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      if (!found)
      {
        /* ------------------------------------------------------ */
        /* try with a .zif directory (if ZipFolders is installed) */
        /* ------------------------------------------------------ */

        sprintf (name, "%s.zif/%s.nv", nvdir, gamename);
        
        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      #endif /* WT_NOUNZIP */

      break;

    case FILETYPE_HIGHSCORE:
      if (mame_highscore_enabled())
      {
        if (!found)
        {
          sprintf (name, "%s/%s.hi", hidir, gamename);

          f->type = kPlainFile;
          f->file = fopen (name, _write ? "wb" : "rb");
          found = f->file != 0;
        }

        #ifndef WT_NOUNZIP

        if (!found)
        {
          /* ---------------------------------------------------- */
          /* try with a .zip directory (if ZipMagic is installed) */
          /* ---------------------------------------------------- */

          sprintf (name, "%s.zip/%s.hi", hidir, gamename);
          
          f->type = kPlainFile;
          f->file = fopen (name, _write ? "wb" : "rb");
          found = f->file != 0;
        }

        if (!found)
        {
          /* ------------------------------------------------------ */
          /* try with a .zif directory (if ZipFolders is installed) */
          /* ------------------------------------------------------ */

          sprintf (name, "%s.zif/%s.hi", hidir, gamename);
          
          f->type = kPlainFile;
          f->file = fopen (name, _write ? "wb" : "rb");
          found = f->file != 0;
        }

        #endif /* WT_NOUNZIP */
      }
      break;

    case FILETYPE_CONFIG:
      
      /* ---------------- */
      /* only for reading */
      /* ---------------- */

      if (_write)
      {
        logerror("osd_fopen: type %02x write not supported\n",filetype);
        break;
      }

      sprintf (name, "%s/%s.cfg", cfgdir, gamename?gamename:filename);
      
      f->type = kPlainFile;
      f->file = fopen (name, _write ? "wb" : "r");
      found = f->file != 0;

      #ifndef WT_NOUNZIP

      if (!found)
      {
        /* ---------------------------------------------------- */
        /* try with a .zip directory (if ZipMagic is installed) */
        /* ---------------------------------------------------- */

        sprintf (name, "%s.zip/%s.cfg", cfgdir, gamename);
        
        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      if (!found)
      {
        /* ------------------------------------------------------ */
        /* try with a .zif directory (if ZipFolders is installed) */
        /* ------------------------------------------------------ */

        sprintf (name, "%s.zif/%s.cfg", cfgdir, gamename);
        
        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      #endif /* WT_NOUNZIP */

      break;

    case FILETYPE_INPUTLOG:
      
      sprintf (name, "%s/%s.inp", inpdir, gamename);
      
      f->type = kPlainFile;
      f->file = fopen (name, _write ? "wb" : "rb");
      found = f->file != 0;

      #ifndef WT_NOUNZIP

      if (!found)
      {
        /* ---------------------------------------------------- */
        /* try with a .zip directory (if ZipMagic is installed) */
        /* ---------------------------------------------------- */

        sprintf (name, "%s.zip/%s.inp", inpdir, gamename);

        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      if (!found)
      {
        /* ------------------------------------------------------ */
        /* try with a .zif directory (if ZipFolders is installed) */
        /* ------------------------------------------------------ */

        sprintf (name, "%s.zif/%s.inp", inpdir, gamename);

        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      if (!_write)
      {
        char file[256];

        sprintf (file, "%s.inp", gamename);
        sprintf (name, "%s/%s.zip", inpdir, gamename);

        LOG(("Trying %s in %s\n", file, name));

        if (!stat(name, &stat_buffer))
        {
          if (!load_zipped_file(name, file, &f->data, &f->length))
          {
            LOG(("Using (osd_fopen) zip file %s for %s\n", name, file));

            f->type = kZippedFile;
            f->offset = 0;
            found = 1;
          }
        }
      }

      #endif /* WT_NOUNZIP */

      break;

    case FILETYPE_STATE:

      sprintf (name, "%s/%s.sta", stadir, gamename);

      f->file = fopen (name, _write ? "wb" : "rb");
      found = !(f->file == 0);

      #ifndef WT_NOUNZIP

      if (!found)
      {
        /* ---------------------------------------------------- */
        /* try with a .zip directory (if ZipMagic is installed) */
        /* ---------------------------------------------------- */

        sprintf (name, "%s.zip/%s.sta", stadir, gamename);

        f->file = fopen (name, _write ? "wb" : "rb");
        found = !(f->file == 0);
      }

      if (!found)
      {
        /* ------------------------------------------------------ */
        /* try with a .zif directory (if ZipFolders is installed) */
        /* ------------------------------------------------------ */

        sprintf (name, "%s.zif/%s.sta", stadir, gamename);

        f->file = fopen (name, _write ? "wb" : "rb");
        found = !(f->file == 0);
      }

      #endif /* WT_NOUNZIP */

      break;

    case FILETYPE_ARTWORK:

      /* ---------------- */
      /* only for reading */
      /* ---------------- */

      if (_write)
      {
        logerror("osd_fopen: type %02x write not supported\n",filetype);
        break;
      }

      sprintf (name, "%s/%s", artworkdir, filename);

      f->type = kPlainFile;
      f->file = fopen (name, _write ? "wb" : "rb");
      found = f->file != 0;

      #ifndef WT_NOUNZIP

      if (!found)
      {
        /* ---------------------------------------------------- */
        /* try with a .zip directory (if ZipMagic is installed) */
        /* ---------------------------------------------------- */

        sprintf (name, "%s.zip/%s.png", artworkdir, filename);

        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      if (!found)
      {
        /* ------------------------------------------------------ */
        /* try with a .zif directory (if ZipFolders is installed) */
        /* ------------------------------------------------------ */

        sprintf (name, "%s.zif/%s.png", artworkdir, filename);

        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      if (!found)
      {
        char file[256], *extension;

        sprintf(file, "%s", filename);
        sprintf(name, "%s/%s", artworkdir, filename);

        extension = strrchr(name, '.');

        if (extension)
          strcpy (extension, ".zip");
        else
          strcat (name, ".zip");

        LOG(("Trying %s in %s\n", file, name));

        if (!stat(name, &stat_buffer))
        {
          if (!load_zipped_file(name, file, &f->data, &f->length))
          {
            LOG(("Using (osd_fopen) zip file %s\n", name));

            f->type = kZippedFile;
            f->offset = 0;
            found = 1;
          }
        }

        if (!found)
        {
          sprintf(name, "%s/%s.zip", artworkdir, game);

          LOG(("Trying %s in %s\n", file, name));

          if (!stat(name, &stat_buffer))
          {
            if (!load_zipped_file (name, file, &f->data, &f->length))
            {
              LOG(("Using (osd_fopen) zip file %s\n", name));

              f->type = kZippedFile;
              f->offset = 0;
              found = 1;
            }
          }
        }
      }

      #endif /* WT_NOUNZIP */

      break;

    case FILETYPE_MEMCARD:

      sprintf (name, "%s/%s", memcarddir, filename);

      f->type = kPlainFile;
      f->file = fopen (name, _write ? "wb" : "rb");
      found = f->file != 0;
      
      break;

    case FILETYPE_SCREENSHOT:
      
      /* ---------------------------------------------------------------- */
      /* Digita, we use snaps in our mamefront UI so we need to read them */
      /* ---------------------------------------------------------------- */

      #if 0
      if (!_write)
      {
        logerror("osd_fopen: type %02x read not supported\n",filetype);
        break;
      }
      #endif
      
      sprintf (name, "%s/%s.png", screenshotdir, filename);
      
      f->type = kPlainFile;
      f->file = fopen (name, _write ? "wb" : "rb");
      found = f->file != 0;

      /* -------------------------------------------------------------------------- */
      /* Digita, the artwork code append the .png so I added this extra file finder */
      /* -------------------------------------------------------------------------- */

      if (!found)
      {
        sprintf (name, "%s/%s", screenshotdir, filename);

        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }

      /* ----------------------------------------------------------------------- */
      /* Dreamcast, if still not found we must be trying to load the splash from */
      /* the built in romdisk area.                                              */
      /* ----------------------------------------------------------------------- */

      if (!found)
      {
        sprintf (name, "/rd/%s", filename);

        f->type = kPlainFile;
        f->file = fopen (name, _write ? "wb" : "rb");
        found = f->file != 0;
      }
      break;

    case FILETYPE_HIGHSCORE_DB:
    case FILETYPE_HISTORY:
      
      /* ---------------- */
      /* only for reading */
      /* ---------------- */

      if (_write)
      {
        logerror("osd_fopen: type %02x write not supported\n",filetype);
        break;
      }

      f->type = kPlainFile;
      
      /* ----------------------------------------------- */
      /* open as ASCII files, not binary like the others */
      /* ----------------------------------------------- */

      f->file = fopen (filename, _write ? "w" : "r");
      found = f->file != 0;
      break;

    case FILETYPE_CHEAT:

      sprintf (name, "%s/%s", cheatdir, filename);

      f->type = kPlainFile;
      
      /* ----------------------------------------------- */
      /* open as ASCII files, not binary like the others */
      /* ----------------------------------------------- */

      f->file = fopen (filename, _write ? "a" : "r");
      found = f->file != 0;
      break;

    case FILETYPE_LANGUAGE:
      
      /* ---------------- */
      /* only for reading */
      /* ---------------- */

      if (_write)
      {
        logerror("osd_fopen: type %02x write not supported\n",filetype);
        break;
      }

      sprintf (name, "%s.lng", filename);

      f->type = kPlainFile;

      /* ----------------------------------------------- */
      /* open as ASCII files, not binary like the others */
      /* ----------------------------------------------- */

      f->file = fopen (name, _write ? "w" : "r");
      found = f->file != 0;
      break;

    default:
      logerror("osd_fopen(): unknown filetype %02x\n",filetype);
  }

  if (!found)
  {
    free (f);
    return 0;
  }

  //printf("osd_fopen (t=%d l=%u)\n",f->type,f->length);

  return f;
}

FILE *get_file_handle(void *file)
{
  if (file)
  {
    FakeFileHandle *f = (FakeFileHandle *) file;

    return(f->file);
  }
  else
    return 0;
}
    

/* osd_fread ==================================================== */
/*                                                                */
/* ============================================================== */

int osd_fread (void *file, void *buffer, int length)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  switch (f->type)
  {
    case kPlainFile:
      return fread (buffer, 1, length, f->file);
      break;

    case kZippedFile:
    case kRAMFile:

      /* ------------------------------------ */
      /* reading from the RAM image of a file */
      /* ------------------------------------ */

      if (f->data)
      {
        if (length + f->offset > f->length)
          length = f->length - f->offset;

        memcpy (buffer, f->offset + f->data, length);

        f->offset += length;

        return length;
      }
      break;
  }

  return 0;
}

/* osd_fread_swap =============================================== */
/*                                                                */
/* ============================================================== */

int osd_fread_swap (void *file, void *buffer, int length)
{
  int i;
  unsigned char *buf;
  unsigned char temp;
  int res;

  res = osd_fread (file, buffer, length);

  buf = buffer;

  for (i = 0; i < length; i += 2)
  {
    temp = buf[i];
    buf[i] = buf[i + 1];
    buf[i + 1] = temp;
  }

  return res;
}

/* osd_fwrite =================================================== */
/*                                                                */
/* ============================================================== */

int osd_fwrite (void *file, const void *buffer, int length)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  switch (f->type)
  {
    case kPlainFile:
      return fwrite (buffer, 1, length, ((FakeFileHandle *) file)->file);
    default:
      return 0;
  }
}

/* osd_fwrite_swap ============================================== */
/*                                                                */
/* ============================================================== */

int osd_fwrite_swap (void *file, const void *buffer, int length)
{
  int i;
  unsigned char *buf;
  unsigned char temp;
  int res;

  buf = (unsigned char *) buffer;

  for (i = 0; i < length; i += 2)
  {
    temp = buf[i];
    buf[i] = buf[i + 1];
    buf[i + 1] = temp;
  }

  res = osd_fwrite (file, buffer, length);

  for (i = 0; i < length; i += 2)
  {
    temp = buf[i];
    buf[i] = buf[i + 1];
    buf[i + 1] = temp;
  }

  return res;
}

#if 0
/* osd_fread_scatter ============================================ */
/*                                                                */
/* ============================================================== */

int osd_fread_scatter (void *file, void *buffer, int length, int increment)
{
  unsigned char *buf = buffer;
  FakeFileHandle *f = (FakeFileHandle *) file;
  unsigned char tempbuf[4096];
  int totread, r, i;

  switch (f->type)
  {
    case kPlainFile:

      totread = 0;

      while (length)
      {
        r = length;

        if (r > 4096)
          r = 4096;

        r = fread (tempbuf, 1, r, f->file);

        if (r == 0)
          return totread;      /* error */

        for (i = 0; i < r; i++)
        {
          *buf = tempbuf[i];
          buf += increment;
        }

        totread += r;

        length -= r;
      }

      return totread;
      break;

    case kZippedFile:
    case kRAMFile:

      /* ------------------------------------ */
      /* reading from the RAM image of a file */
      /* ------------------------------------ */

      if (f->data)
      {
        if (length + f->offset > f->length)
          length = f->length - f->offset;

        for (i = 0; i < length; i++)
        {
          *buf = f->data[f->offset + i];
          buf += increment;
        }

        f->offset += length;

        return length;
      }
      break;
  }

  return 0;
}
#endif

/* osd_fseek ==================================================== */
/*                                                                */
/* ============================================================== */

int osd_fseek (void *file, int offset, int whence)
{
  FakeFileHandle *f = (FakeFileHandle *) file;
  int err = 0;

  switch (f->type)
  {
    case kPlainFile:

      return fseek (f->file, offset, whence);
      break;

    case kZippedFile:
    case kRAMFile:
      
      /* -------------------------------------- */
      /* seeking within the RAM image of a file */
      /* -------------------------------------- */

      switch (whence)
      {
        case SEEK_SET:
          f->offset = offset;
          break;
        case SEEK_CUR:
          f->offset += offset;
          break;
        case SEEK_END:
          f->offset = f->length + offset;
          break;
      }
      break;
  }

  return err;
}

/* osd_fclose =================================================== */
/*                                                                */
/* ============================================================== */

void osd_fclose (void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  //printf("osd_fclose (t=%d l=%u)\n",f->type,f->length);

  switch (f->type)
  {
    case kPlainFile:
      fclose (f->file);
      break;
    case kZippedFile:
    case kRAMFile:
      if (f->data)
        free (f->data);
      break;
  }
  free (f);
}

/* checksum_file ================================================ */
/*                                                                */
/* ============================================================== */

static int checksum_file (const char *file, unsigned char **p, unsigned int *size, unsigned int *crc)
{
  int length;
  unsigned char *data;
  FILE *f;

  f = fopen (file, "rb");
  if (!f)
    return -1;

  /* determine length of file */
  if (fseek (f, 0L, SEEK_END) != 0)
  {
    fclose (f);
    return -1;
  }

  length = ftell (f);
  if (length == -1L)
  {
    fclose (f);
    return -1;
  }

  /* allocate space for entire file */
  data = (unsigned char *) malloc (length);
  if (!data)
  {
    fclose (f);
    return -1;
  }

  /* read entire file into memory */
  if (fseek (f, 0L, SEEK_SET) != 0)
  {
    free (data);
    fclose (f);
    return -1;
  }

  if (fread (data, sizeof (unsigned char), length, f) != length)
  {
    free (data);
    fclose (f);
    return -1;
  }

  *size = length;
  *crc = crc32 (0L, data, length);
  if (p)
    *p = data;
  else
    free (data);

  fclose (f);

  return 0;
}

/* osd_fchecksum ================================================ */
/*                                                                */
/* ============================================================== */

int osd_fchecksum (const char *game, const char *filename, unsigned int *length, unsigned int *sum)
{
  char name[256];
  int indx;
  struct mame_stat stat_buffer;
  int found = 0;
  const char *gamename = game;

  /* Support "-romdir" yuck. */
  if (alternate_name)
    gamename = alternate_name;

  for (indx = 0; indx < rompathc && !found; ++indx)
  {
    const char *dir_name = rompathv[indx];

    if (!found)
    {
      sprintf (name, "%s/%s", dir_name, gamename);

      if (stat_dir(name, &stat_buffer) == 0 && (stat_buffer.st_mode & S_IFDIR))
      {
        sprintf (name, "%s/%s/%s", dir_name, gamename, filename);

        if (checksum_file (name, 0, length, sum) == 0)
        {
          found = 1;
        }
      }
    }

    #ifndef WT_NOUNZIP

    if (!found)
    {
      /* ------------------------- */
      /* try with a .zip extension */
      /* ------------------------- */

      sprintf (name, "%s/%s.zip", dir_name, gamename);

      if (!stat(name, &stat_buffer))
      {
        if (!checksum_zipped_file(name, filename, length, sum))
        {
          LOG(("Using (osd_fchecksum) zip file for %s\n", filename));
          found = 1;
        }
      }
    }

    if (!found)
    {
      /* ------------------------------------------------------ */
      /* try with a .zif directory (if ZipFolders is installed) */
      /* ------------------------------------------------------ */

      sprintf (name, "%s/%s.zif", dir_name, gamename);

      if (!stat_dir(name, &stat_buffer))
      {
        sprintf (name, "%s/%s.zif/%s", dir_name, gamename, filename);

        if (!checksum_file (name, 0, length, sum))
        {
          found = 1;
        }
      }
    }

    #endif /* WT_NOUNZIP */
  }

  if (!found)
    return -1;

  return 0;
}

/* osd_fsize ==================================================== */
/*                                                                */
/* ============================================================== */

int osd_fsize (void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  if (f->type == kRAMFile || f->type == kZippedFile)
    return f->length;

  if (f->file)
  {
    int size, offs;
    offs = ftell( f->file );
    fseek( f->file, 0, SEEK_END );
    size = ftell( f->file );
    fseek( f->file, offs, SEEK_SET );
    return size;
  }

  return 0;
}

/* osd_fcrc ===================================================== */
/*                                                                */
/* ============================================================== */

unsigned int osd_fcrc (void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  return f->crc;
}

/* osd_fgetc ==================================================== */
/*                                                                */
/* ============================================================== */

int osd_fgetc(void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  if (f->type == kPlainFile && f->file)
    return fgetc(f->file);
  else
    return EOF;
}

/* osd_ungetc =================================================== */
/*                                                                */
/* ============================================================== */

int osd_ungetc(int c, void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  if (f->type == kPlainFile && f->file)
    return ungetc(c,f->file);
  else
    return EOF;
}

/* osd_fgets ==================================================== */
/*                                                                */
/* ============================================================== */

char *osd_fgets(char *s, int n, void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  if (f->type == kPlainFile && f->file)
  {
    char *ret;

    ret = 0;

    if (n)
    {
      *s = 0;
      ret = fgets(s,n,f->file);
      if (ret)
      {
        s[n - 1] = 0;
        n = strlen(s);
        if (n)
        {
          // cut off cr-lf at end
          if (s[n-1] < 32)
          {
            s[n-1] = 0;
          }
          if (s[n-2] < 32)
          {
            s[n-2] = 0;
          }
        }
      }
    }
    return ret;
  }
  else
    return NULL;
}

/* osd_feof ===================================================== */
/*                                                                */
/* ============================================================== */

int osd_feof(void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  if (f->type == kPlainFile && f->file)
    return feof(f->file);
  else
    return 1;
}

/* osd_ftell ==================================================== */
/*                                                                */
/* ============================================================== */

int osd_ftell(void *file)
{
  FakeFileHandle *f = (FakeFileHandle *) file;

  if (f->type == kPlainFile && f->file)
    return ftell(f->file);
  else
    return -1L;
}


/* osd_display_loading_rom_message ============================== */
/*                                                                */
/* Called while loading ROMs. It is called a last time with name  */
/* == 0 to signal that the ROM loading process is finished.       */
/* Return non-zero to abort loading.                              */
/*                                                                */
/* ============================================================== */

int osd_display_loading_rom_message (const char *name, int current, int total)
{
  /* ---------------------------- */
  /* Digita, we don't have stdout */
  /* ---------------------------- */

  if (name)
    printf ("loading %s\n", name);
  else
    malloc_stats();

  return 0;
}

#endif  /* !MESS */


