// MAME includes
#include "dreamcast.h"
#include "driver.h"

//#define USER1START			profiler_mark(PROFILER_USER1);
//#define USER1END(ret)		{ profiler_mark(PROFILER_END); return ret; }

#include <kos.h>

void poll_joysticks(uint32 code);

unsigned long kp_count = 0;
unsigned long js_calls = 0;

/* ------------------------------------------ */
/* NOTE: 5th joystick addr is just a sentinel */
/* ------------------------------------------ */

maple_device_t *maple_keyboard_device = 0;
maple_device_t *maple_controller_dev[MAPLE_PORT_COUNT+1] = {0,0,0,0,0};
unsigned char   controller_type[MAPLE_PORT_COUNT] = {0,0,0,0};

kbd_state_t  def_kbd_state;
kbd_state_t  *kbd_state = &def_kbd_state;

cont_state_t def_cont_state;
cont_state_t *cont_state[MAPLE_PORT_COUNT] = {&def_cont_state,&def_cont_state,&def_cont_state,&def_cont_state};

unsigned char max_players = 0;
unsigned char ui_mode     = 0;

#define KBD_KEY_LSHIFT 		0x70
#define KBD_KEY_RSHIFT 		0x71
#define KBD_KEY_LCONTROL 	0x72
#define KBD_KEY_RCONTROL 	0x73
#define KBD_KEY_ALT 		  0x74
#define KBD_KEY_ALTGR 		0x75

static struct KeyboardInfo keylist[] =
{
	{ "A",			KBD_KEY_A,				KEYCODE_A },
	{ "B",			KBD_KEY_B,				KEYCODE_B },
	{ "C",			KBD_KEY_C,				KEYCODE_C },
	{ "D",			KBD_KEY_D,				KEYCODE_D },
	{ "E",			KBD_KEY_E,				KEYCODE_E },
	{ "F",			KBD_KEY_F,				KEYCODE_F },
	{ "G",			KBD_KEY_G,				KEYCODE_G },
	{ "H",			KBD_KEY_H,				KEYCODE_H },
	{ "I",			KBD_KEY_I,				KEYCODE_I },
	{ "J",			KBD_KEY_J,				KEYCODE_J },
	{ "K",			KBD_KEY_K,				KEYCODE_K },
	{ "L",			KBD_KEY_L,				KEYCODE_L },
	{ "M",			KBD_KEY_M,				KEYCODE_M },
	{ "N",			KBD_KEY_N,				KEYCODE_N },
	{ "O",			KBD_KEY_O,				KEYCODE_O },
	{ "P",			KBD_KEY_P,				KEYCODE_P },
	{ "Q",			KBD_KEY_Q,				KEYCODE_Q },
	{ "R",			KBD_KEY_R,				KEYCODE_R },
	{ "S",			KBD_KEY_S,				KEYCODE_S },
	{ "T",			KBD_KEY_T,				KEYCODE_T },
	{ "U",			KBD_KEY_U,				KEYCODE_U },
	{ "V",			KBD_KEY_V,				KEYCODE_V },
	{ "W",			KBD_KEY_W,				KEYCODE_W },
	{ "X",			KBD_KEY_X,				KEYCODE_X },
	{ "Y",			KBD_KEY_Y,				KEYCODE_Y },
	{ "Z",			KBD_KEY_Z,				KEYCODE_Z },
	{ "0",			KBD_KEY_0,				KEYCODE_0 },
	{ "1",			KBD_KEY_1,				KEYCODE_1 },
	{ "2",			KBD_KEY_2,				KEYCODE_2 },
	{ "3",			KBD_KEY_3,				KEYCODE_3 },
	{ "4",			KBD_KEY_4,				KEYCODE_4 },
	{ "5",			KBD_KEY_5,				KEYCODE_5 },
	{ "6",			KBD_KEY_6,				KEYCODE_6 },
	{ "7",			KBD_KEY_7,				KEYCODE_7 },
	{ "8",			KBD_KEY_8,				KEYCODE_8 },
	{ "9",			KBD_KEY_9,				KEYCODE_9 },
	{ "0 PAD",		KBD_KEY_PAD_0,			KEYCODE_0_PAD },
	{ "1 PAD",		KBD_KEY_PAD_1,			KEYCODE_1_PAD },
	{ "2 PAD",		KBD_KEY_PAD_2,			KEYCODE_2_PAD },
	{ "3 PAD",		KBD_KEY_PAD_3,			KEYCODE_3_PAD },
	{ "4 PAD",		KBD_KEY_PAD_4,			KEYCODE_4_PAD },
	{ "5 PAD",		KBD_KEY_PAD_5,			KEYCODE_5_PAD },
	{ "6 PAD",		KBD_KEY_PAD_6,			KEYCODE_6_PAD },
	{ "7 PAD",		KBD_KEY_PAD_7,			KEYCODE_7_PAD },
	{ "8 PAD",		KBD_KEY_PAD_8,			KEYCODE_8_PAD },
	{ "9 PAD",		KBD_KEY_PAD_9,			KEYCODE_9_PAD },
	{ "F1",			KBD_KEY_F1,				KEYCODE_F1 },
	{ "F2",			KBD_KEY_F2,				KEYCODE_F2 },
	{ "F3",			KBD_KEY_F3,				KEYCODE_F3 },
	{ "F4",			KBD_KEY_F4,				KEYCODE_F4 },
	{ "F5",			KBD_KEY_F5,				KEYCODE_F5 },
	{ "F6",			KBD_KEY_F6,				KEYCODE_F6 },
	{ "F7",			KBD_KEY_F7,				KEYCODE_F7 },
	{ "F8",			KBD_KEY_F8,				KEYCODE_F8 },
	{ "F9",			KBD_KEY_F9,				KEYCODE_F9 },
	{ "F10",		KBD_KEY_F10,			KEYCODE_F10 },
	{ "F11",		KBD_KEY_F11,			KEYCODE_F11 },
	{ "F12",		KBD_KEY_F12,			KEYCODE_F12 },
	{ "ESC",		KBD_KEY_ESCAPE,			KEYCODE_ESC },
	{ "~",			KBD_KEY_TILDE,			KEYCODE_TILDE },
	{ "-",          KBD_KEY_MINUS,          KEYCODE_MINUS },
//	{ "=",          KBD_KEY_EQUALS,         KEYCODE_EQUALS },
	{ "BKSPACE",	KBD_KEY_BACKSPACE,		KEYCODE_BACKSPACE },
	{ "TAB",		KBD_KEY_TAB,			KEYCODE_TAB },
	{ "[",          KBD_KEY_LBRACKET,      KEYCODE_OPENBRACE },
	{ "]",          KBD_KEY_RBRACKET,     KEYCODE_CLOSEBRACE },
	{ "ENTER",		KBD_KEY_ENTER,			KEYCODE_ENTER },
	{ ";",          KBD_KEY_SEMICOLON,          KEYCODE_COLON },
	{ ":",          KBD_KEY_QUOTE,          KEYCODE_QUOTE },
	{ "\\",         KBD_KEY_BACKSLASH,      KEYCODE_BACKSLASH },
//	{ "<",          KBD_KEY_BACKSLASH2,     KEYCODE_BACKSLASH2 },
	{ ",",          KBD_KEY_COMMA,          KEYCODE_COMMA },
	{ ".",          KBD_KEY_PERIOD,           KEYCODE_STOP },
	{ "/",          KBD_KEY_SLASH,          KEYCODE_SLASH },
	{ "SPACE",		KBD_KEY_SPACE,			KEYCODE_SPACE },
	{ "INS",		KBD_KEY_INSERT,			KEYCODE_INSERT },
	{ "DEL",		KBD_KEY_DEL,			KEYCODE_DEL },
	{ "HOME",		KBD_KEY_HOME,			KEYCODE_HOME },
	{ "END",		KBD_KEY_END,			KEYCODE_END },
	{ "PGUP",		KBD_KEY_PGUP,			KEYCODE_PGUP },
	{ "PGDN",		KBD_KEY_PGDOWN,			KEYCODE_PGDN },
	{ "LEFT",		KBD_KEY_LEFT,			KEYCODE_LEFT },
	{ "RIGHT",		KBD_KEY_RIGHT,			KEYCODE_RIGHT },
	{ "UP",			KBD_KEY_UP,				KEYCODE_UP },
	{ "DOWN",		KBD_KEY_DOWN,			KEYCODE_DOWN },
	{ "/ PAD",      KBD_KEY_PAD_DIVIDE,      KEYCODE_SLASH_PAD },
	{ "* PAD",      KBD_KEY_PAD_MULTIPLY,       KEYCODE_ASTERISK },
	{ "- PAD",      KBD_KEY_PAD_MINUS,      KEYCODE_MINUS_PAD },
	{ "+ PAD",      KBD_KEY_PAD_PLUS,       KEYCODE_PLUS_PAD },
	{ ". PAD",      KBD_KEY_PAD_PERIOD,        KEYCODE_DEL_PAD },
	{ "ENTER PAD",  KBD_KEY_PAD_ENTER,      KEYCODE_ENTER_PAD },
	{ "PRTSCR",     KBD_KEY_PRINT,         KEYCODE_PRTSCR },
	{ "PAUSE",      KBD_KEY_PAUSE,          KEYCODE_PAUSE },
	{ "LSHIFT",		KBD_KEY_LSHIFT,			KEYCODE_LSHIFT },
	{ "RSHIFT",		KBD_KEY_RSHIFT,			KEYCODE_RSHIFT },
	{ "LCTRL",		KBD_KEY_LCONTROL,		KEYCODE_LCONTROL },
	{ "RCTRL",		KBD_KEY_RCONTROL,		KEYCODE_RCONTROL },
	{ "ALT",		KBD_KEY_ALT,			KEYCODE_LALT },
	{ "ALTGR",		KBD_KEY_ALTGR,			KEYCODE_RALT },
//	{ "LWIN",		KBD_KEY_LWIN,			KEYCODE_OTHER },
//	{ "RWIN",		KBD_KEY_RWIN,			KEYCODE_OTHER },
//	{ "MENU",		KBD_KEY_MENU,			KEYCODE_OTHER },
	{ "SCRLOCK",    KBD_KEY_SCRLOCK,        KEYCODE_SCRLOCK },
	{ "NUMLOCK",    KBD_KEY_PAD_NUMLOCK,        KEYCODE_NUMLOCK },
	{ "CAPSLOCK",   KBD_KEY_CAPSLOCK,       KEYCODE_CAPSLOCK },
	{ 0, 0, 0 }	/* end of table */
};


/* return a list of all available keys */
const struct KeyboardInfo *osd_get_key_list(void)
{
	return keylist;
}

void osd_led_w(int led, int on)
{
}

/* check if a key is pressed. The keycode is the standard PC keyboard code, as */
/* defined in osdepend.h. Return 0 if the key is not pressed, nonzero otherwise. */
int osd_is_key_pressed(int keycode)
{	
  //js_calls++;

  /* -------- */
  /* KEYBOARD */
  /* -------- */

  #if !defined(WT_NOKEYBOARD) //OVERCLOCK_STUFF
  if (maple_keyboard_device &&
      (kbd_state->cond.keys[0] == keycode ||
       kbd_state->cond.keys[1] == keycode ||
       kbd_state->cond.keys[2] == keycode ||
       kbd_state->cond.keys[3] == keycode ||
       kbd_state->cond.keys[4] == keycode ||
       kbd_state->cond.keys[5] == keycode))
    return 1;
  #endif
  
  /* ----------- */
  /* CONTROLLERS */
  /* ----------- */

  #define ADMIN_CONT      (CONT_DPAD_UP | CONT_DPAD_DOWN)
  #define ADMIN_CONT_COIN (CONT_START | CONT_DPAD_UP)
  #define ADMIN_CONT_MENU (CONT_START | CONT_DPAD_DOWN)

  switch(keycode) 
  {
    case KBD_KEY_ESCAPE:
      if ((cont_state[0]->buttons & CONT_A) && (cont_state[0]->buttons & CONT_START))
      {
        do
        {
          poll_joysticks(1);
          timer_spin_sleep(50);
        }
        while ((cont_state[0]->buttons & CONT_A) || (cont_state[0]->buttons & CONT_START));

        return 1;
      }
      else return 0;

    case KBD_KEY_PAUSE:
      if ((cont_state[0]->buttons & CONT_Y) && (cont_state[0]->buttons & CONT_START))
      {
        do
        {
          poll_joysticks(1);
          timer_spin_sleep(50);
        }
        while ((cont_state[0]->buttons & CONT_Y) || (cont_state[0]->buttons & CONT_START));

        return 1;
      }
      else return 0;

    case KBD_KEY_ALTGR:
      return(kbd_state->cond.modifiers & KBD_MOD_RALT);
    case KBD_KEY_5:
      return((cont_state[0]->buttons & ADMIN_CONT_COIN) == ADMIN_CONT_COIN);                      /* COIN 1 */
    case KBD_KEY_1:
      return((cont_state[0]->buttons & CONT_START) && !(cont_state[0]->buttons & ADMIN_CONT));    /* P1 START */
    case KBD_KEY_UP:
      return(cont_state[0]->buttons & CONT_DPAD_UP);                                              /* UP */
    case KBD_KEY_DOWN:
      return(cont_state[0]->buttons & CONT_DPAD_DOWN);                                            /* DOWN */
    case KBD_KEY_LEFT:
      return(cont_state[0]->buttons & CONT_DPAD_LEFT);                                            /* LEFT */
    case KBD_KEY_RIGHT:
      return(cont_state[0]->buttons & CONT_DPAD_RIGHT);                                           /* RIGHT */
    case KBD_KEY_ENTER:
    case KBD_KEY_LCONTROL:
      return((kbd_state->cond.modifiers & KBD_MOD_LCTRL) || (cont_state[0]->buttons & CONT_A));   /* UI SELECT & BUTTON 1 */
    case KBD_KEY_ALT:
      return((kbd_state->cond.modifiers & KBD_MOD_LALT) || (cont_state[0]->buttons & CONT_B));    /* BUTTON 2 */
    case KBD_KEY_SPACE:
      return(cont_state[0]->buttons & CONT_X);                                                    /* BUTTON 3 */
    case KBD_KEY_LSHIFT:
      return((kbd_state->cond.modifiers & KBD_MOD_LSHIFT) || (cont_state[0]->buttons & CONT_Y));  /* BUTTON 4 */
    case KBD_KEY_TAB:
      return((cont_state[0]->buttons & ADMIN_CONT_MENU) == ADMIN_CONT_MENU);                      /* UI TOGGLE */
    case KBD_KEY_6:
      return((cont_state[1]->buttons & ADMIN_CONT_COIN) == ADMIN_CONT_COIN);                      /* COIN 2 */
    case KBD_KEY_2:
      return((cont_state[1]->buttons & CONT_START) && !(cont_state[1]->buttons & ADMIN_CONT));    /* P2 START */
    case KBD_KEY_R:
      return(cont_state[1]->buttons & CONT_DPAD_UP);                                              /* UP */    
    case KBD_KEY_F:                                                                                           
      return(cont_state[1]->buttons & CONT_DPAD_DOWN);                                            /* DOWN */  
    case KBD_KEY_D:                                                                                           
      return(cont_state[1]->buttons & CONT_DPAD_LEFT);                                            /* LEFT */  
    case KBD_KEY_G:                                                                                           
      return(cont_state[1]->buttons & CONT_DPAD_RIGHT);                                           /* RIGHT */ 
    case KBD_KEY_A:
      return(cont_state[1]->buttons & CONT_A);                                                    /* BUTTON 1 */
    case KBD_KEY_S:
      return(cont_state[1]->buttons & CONT_B);                                                    /* BUTTON 2 */
    case KBD_KEY_Q:
      return(cont_state[1]->buttons & CONT_X);                                                    /* BUTTON 3 */
    case KBD_KEY_W:
      return(cont_state[1]->buttons & CONT_Y);                                                    /* BUTTON 4 */
    case KBD_KEY_7:
      return((cont_state[2]->buttons & ADMIN_CONT_COIN) == ADMIN_CONT_COIN);                      /* COIN 3 */
    case KBD_KEY_3:
      return((cont_state[2]->buttons & CONT_START) && !(cont_state[2]->buttons & ADMIN_CONT));    /* P3 START */
    case KBD_KEY_I:
      return(cont_state[2]->buttons & CONT_DPAD_UP);                                              /* UP */   
    case KBD_KEY_K:                                                                                          
      return(cont_state[2]->buttons & CONT_DPAD_DOWN);                                            /* DOWN */ 
    case KBD_KEY_J:                                                                                          
      return(cont_state[2]->buttons & CONT_DPAD_LEFT);                                            /* LEFT */ 
    case KBD_KEY_L:                                                                                          
      return(cont_state[2]->buttons & CONT_DPAD_RIGHT);                                           /* RIGHT */
    case KBD_KEY_RCONTROL:
      return((kbd_state->cond.modifiers & KBD_MOD_RCTRL) || (cont_state[2]->buttons & CONT_A));   /* BUTTON 1 */
    case KBD_KEY_RSHIFT:
      return((kbd_state->cond.modifiers & KBD_MOD_RSHIFT) || (cont_state[2]->buttons & CONT_B));  /* BUTTON 2 */
    case KBD_KEY_PERIOD:
      return(cont_state[2]->buttons & CONT_X);                                                    /* BUTTON 3 */
    case KBD_KEY_SLASH:
      return(cont_state[2]->buttons & CONT_Y);                                                    /* BUTTON 4 */
    case KBD_KEY_8:
      return((cont_state[3]->buttons & ADMIN_CONT_COIN) == ADMIN_CONT_COIN);                      /* COIN 4 */
    case KBD_KEY_4:
      return((cont_state[3]->buttons & CONT_START) && !(cont_state[3]->buttons & ADMIN_CONT));    /* P4 START */
    case KBD_KEY_SEMICOLON:
      return(cont_state[3]->buttons & CONT_DPAD_UP);                                              /* UP */   
    case KBD_KEY_QUOTE:                                                                                      
      return(cont_state[3]->buttons & CONT_DPAD_DOWN);                                            /* DOWN */ 
    case KBD_KEY_BACKSLASH:                                                                                  
      return(cont_state[3]->buttons & CONT_DPAD_LEFT);                                            /* LEFT */ 
    case KBD_KEY_COMMA:                                                                                      
      return(cont_state[3]->buttons & CONT_DPAD_RIGHT);                                           /* RIGHT */
    case KBD_KEY_F1:
      return(cont_state[3]->buttons & CONT_A);                                                    /* BUTTON 1 */
    case KBD_KEY_F2:
      return(cont_state[3]->buttons & CONT_B);                                                    /* BUTTON 2 */
    case KBD_KEY_LBRACKET:
      return(cont_state[3]->buttons & CONT_X);                                                    /* BUTTON 3 */
    case KBD_KEY_RBRACKET:
      return(cont_state[3]->buttons & CONT_Y);                                                    /* BUTTON 4 */
	}

	return 0;
}

int osd_readkey_unicode(int flush)
{
	return 0;
}

/* wait for a key press and return the keycode */
int osd_wait_keypress(void)
{
	return 0;
} 

#define JOYCODE(joy,stick,axis_or_button,dir) \
		((((dir)&0x03)<<14)|(((axis_or_button)&0x3f)<<8)|(((stick)&0x1f)<<3)|(((joy)&0x07)<<0))

#define GET_JOYCODE_JOY(code) (((code)>>0)&0x07)
#define GET_JOYCODE_BUTTON(code) (((code)>>8)&0x3f)

#define MAX_JOY 256
#define MAX_JOY_NAME_LEN 40

static int last_check = -1;

#if 0
static int joyequiv[][2] =
{
	{ JOYCODE(1,0,1,0),	JOYCODE_1_BUTTON1 },
	{ JOYCODE(1,0,2,0),	JOYCODE_1_BUTTON2 },
	{ JOYCODE(1,0,3,0),	JOYCODE_1_BUTTON3 },
	{ JOYCODE(1,0,4,0),	JOYCODE_1_BUTTON4 },
	{ JOYCODE(1,0,5,0),	JOYCODE_1_BUTTON5 },
	{ JOYCODE(1,0,6,0),	JOYCODE_1_LEFT },
	{ JOYCODE(1,0,7,0),	JOYCODE_1_RIGHT },
	{ JOYCODE(1,0,8,0),	JOYCODE_1_UP },
	{ JOYCODE(1,0,9,0),	JOYCODE_1_DOWN },
	{ JOYCODE(1,0,10,0),JOYCODE_1_BUTTON6 },
	
	{ JOYCODE(2,0,1,0),	JOYCODE_2_BUTTON1 },
	{ JOYCODE(2,0,2,0),	JOYCODE_2_BUTTON2 },
	{ JOYCODE(2,0,3,0),	JOYCODE_2_BUTTON3 },
	{ JOYCODE(2,0,4,0),	JOYCODE_2_BUTTON4 },
	{ JOYCODE(2,0,5,0),	JOYCODE_2_BUTTON5 },
	{ JOYCODE(2,0,6,0),	JOYCODE_2_LEFT },
	{ JOYCODE(2,0,7,0),	JOYCODE_2_RIGHT },
	{ JOYCODE(2,0,8,0),	JOYCODE_2_UP },
	{ JOYCODE(2,0,9,0),	JOYCODE_2_DOWN },
	{ JOYCODE(2,0,10,0),JOYCODE_2_BUTTON6 },

	{ JOYCODE(3,0,1,0),	JOYCODE_3_BUTTON1 },
	{ JOYCODE(3,0,2,0),	JOYCODE_3_BUTTON2 },
	{ JOYCODE(3,0,3,0),	JOYCODE_3_BUTTON3 },
	{ JOYCODE(3,0,4,0),	JOYCODE_3_BUTTON4 },
	{ JOYCODE(3,0,5,0),	JOYCODE_3_BUTTON5 },
	{ JOYCODE(3,0,6,0),	JOYCODE_3_LEFT },
	{ JOYCODE(3,0,7,0),	JOYCODE_3_RIGHT },
	{ JOYCODE(3,0,8,0),	JOYCODE_3_UP },
	{ JOYCODE(3,0,9,0),	JOYCODE_3_DOWN },
	{ JOYCODE(3,0,10,0),JOYCODE_3_BUTTON6 },

	{ JOYCODE(4,0,1,0),	JOYCODE_4_BUTTON1 },
	{ JOYCODE(4,0,2,0),	JOYCODE_4_BUTTON2 },
	{ JOYCODE(4,0,3,0),	JOYCODE_4_BUTTON3 },
	{ JOYCODE(4,0,4,0),	JOYCODE_4_BUTTON4 },
	{ JOYCODE(4,0,5,0),	JOYCODE_4_BUTTON5 },
	{ JOYCODE(4,0,6,0),	JOYCODE_4_LEFT },
	{ JOYCODE(4,0,7,0),	JOYCODE_4_RIGHT },
	{ JOYCODE(4,0,8,0),	JOYCODE_4_UP },
	{ JOYCODE(4,0,9,0),	JOYCODE_4_DOWN },
	{ JOYCODE(4,0,10,0),JOYCODE_4_BUTTON6 },
	{ 0,0 }
};

static char joynames[MAX_JOY][MAX_JOY_NAME_LEN+1];	/* will be used to store names for the above */

static char joy_button_name[11][MAX_JOY_NAME_LEN + 1] = 
{
	"A",
	"B",
	"X",
	"Y",
	"Start",
	"Left",
	"Right",
	"Up",
	"Down",
	"R Trig",
	"L Trig"
};

static int joy_button_mask[9] = 
{
	CONT_A,
	CONT_B,
  	CONT_X,
  	CONT_Y,
  	CONT_START,
  	CONT_DPAD_LEFT,
  	CONT_DPAD_RIGHT,
  	CONT_DPAD_UP,
  	CONT_DPAD_DOWN
};
#endif

static struct JoystickInfo joylist[MAX_JOY] =
{
	/* will be filled later */
	{ 0, 0, 0 }	/* end of table */
};

static void init_joy_list(void)
{
  // sorry but don't even think about adding real joysticks, it will slow it way down real fast

  #if 0
	int tot,i,j;
	char buf[256];

	tot = 0;

	// four dreamcast controllers
	for (i = 0;i < count;i++)
	{		
		// Start, A, B, X, & Y, up, down, left, right, right & left trigger
		for (j = 0;j < 11;j++)
		{
			sprintf(buf,"J%d %s",i+1,joy_button_name[j]);
			strncpy(joynames[tot],buf,MAX_JOY_NAME_LEN);
			joynames[tot][MAX_JOY_NAME_LEN] = 0;
			joylist[tot].name = joynames[tot];
			joylist[tot].code = JOYCODE(i+1,0,j+1,0);
			tot++;
		}
	}

	/* terminate array */
	joylist[tot].name = 0;
	joylist[tot].code = 0;
	joylist[tot].standardcode = 0;

	/* fill in equivalences */
	for (i = 0;i < tot;i++)
	{
		joylist[i].standardcode = JOYCODE_OTHER;

		j = 0;
		while (joyequiv[j][0] != 0)
		{
			if (joyequiv[j][0] == joylist[i].code)
			{
				joylist[i].standardcode = joyequiv[j][1];
				break;
			}
			j++;
		}
	}
  #endif
}

/* return a list of all available joys */
const struct JoystickInfo *osd_get_joy_list(void)
{
//	init_joy_list();
	
	return joylist;
}

int osd_is_joy_pressed(int joycode)
{
  // sorry but don't even think about adding real joysticks, it will slow it way down real fast
	return 0;

  #if 0
	int joy_num,button;


	// check for max joysticks
	joy_num = GET_JOYCODE_JOY(joycode) - 1;
	if (joy_num > 3)
	{
		return 0;
	}
	
	// no joystick
	if (maple_controller_dev[joy_num] == 0)
	{
		return 0;
	}

	button = GET_JOYCODE_BUTTON(joycode) - 1;

	if (button == 9)
	{
		return (cont_state[joy_num] && cont_state[joy_num]->rtrig > 16);
	}
	else if (button == 10)
	{
		return (cont_state[joy_num] && cont_state[joy_num]->ltrig > 16);
	}
	else
	{
		return (cont_state[joy_num] && cont_state[joy_num]->buttons & joy_button_mask[button]);
	}
  #endif
}

void poll_joysticks(uint32 code)
{
  WT_PENTRY(1);

  /* --------------------------------------------- */
  /* Only check if we have not checked this vblank */
  /* --------------------------------------------- */

	if (last_check != maple_state.vbl_cntr)
  {
    int count = 0;

	  last_check = maple_state.vbl_cntr;

    if (maple_keyboard_device && !((kbd_state = (kbd_state_t*)maple_dev_status(maple_keyboard_device))))
      kbd_state = &def_kbd_state;

	  while(maple_controller_dev[count])
    {
		  if (!((cont_state[count] = (cont_state_t *)maple_dev_status(maple_controller_dev[count]))))
        cont_state[count] = &def_cont_state;

      count++;
    }
  }

  WT_PEXIT(1);

  return;
}

int osd_joystick_needs_calibration (void)
{
	return 0;
}

void osd_joystick_start_calibration (void)
{
}

const char *osd_joystick_calibrate_next (void)
{
	return 0;
}

void osd_joystick_calibrate (void)
{
}

void osd_joystick_end_calibration (void)
{
}

int osd_is_joystick_axis_code(int joycode)
{
  return(0);
}

/* return a value in the range -128 .. 128 (yes, 128, not 127) */
void osd_analogjoy_read(int player,int analog_axis[MAX_ANALOG_AXES], InputCode analogjoy_input[MAX_ANALOG_AXES])
{
  #if 0
  switch(controller_type[player])
  {
    case 1:  /* controller */
    case 4:  /* lightgun */
      *analog_x = cont_state[player]->joyx;
      *analog_y = cont_state[player]->joyy;
      break;

    case 2:  /* steering wheel  */
      *analog_x = cont_state[player]->joyx;
      *analog_y = ((cont_state[player]->rtrig>cont_state[player]->ltrig)?cont_state[player]->rtrig/2:-1*(cont_state[player]->ltrig/2));
      break;

    default: /* other */
      *analog_x = *analog_y = 0;
      break;
  }
  #endif
  
  return;
}

void osd_lightgun_read(int player,int *deltax,int *deltay)
{
  return(osd_analogjoy_read(player,deltax,deltay));
}

void osd_trak_read(int player,int *deltax,int *deltay)
{
  switch(controller_type[player])
  {
    case 2: /* steering wheel */
      *deltay = 
        ((cont_state[player]->rtrig>cont_state[player]->ltrig)?
         cont_state[player]->rtrig/2:-1*(cont_state[player]->ltrig/2));
      break;

    case 8: /* mouse */
      {
        mouse_state_t *mouse_state = (mouse_state_t *)cont_state[player];

        *deltax = mouse_state->dx;
        *deltay = mouse_state->dy;
      }
      break;

    default: /* other */
      *deltax = cont_state[player]->joyx;
      *deltay = cont_state[player]->joyy;
      break;
  }

  return;
}

void osd_customize_inputport_defaults(struct ipd *defaults)
{
	while (defaults->type != IPT_END)
	{
    UINT32 player = (defaults->type & IPF_PLAYERMASK);

    /* ----------------------------------------------------------- */
    /* Clear the sequence settings for any non-connected joysticks */
    /* ----------------------------------------------------------- */

    if (player && !maple_controller_dev[player>>16])
    {
      //printf("Cleared: %d '%s' (player %lu) mp: %d\n",defaults->type,defaults->name,player>>16,(int)max_players);
      seq_set_0(&defaults->seq);
    }
    else
    {
      switch(player>>16)
      {
        case 0: /* PLAYER1 & DEFAULTS */
        {
          switch(defaults->type & ~IPF_PLAYERMASK)
          {
            /* --- */
            /* SET */
            /* --- */

            case IPT_JOYSTICKRIGHT_UP:
              if (ui_mode > 0 && max_players > 1)
                seq_set_1(&defaults->seq,KEYCODE_R);
              else
                seq_set_1(&defaults->seq,KEYCODE_LSHIFT);
              break;
            case IPT_JOYSTICKRIGHT_DOWN:
              if (ui_mode > 0 && max_players > 1)
                seq_set_1(&defaults->seq,KEYCODE_F);
              else
                seq_set_1(&defaults->seq,KEYCODE_LCONTROL);
              break;
            case IPT_JOYSTICKRIGHT_LEFT:
              if (ui_mode > 0 && max_players > 1)
                seq_set_1(&defaults->seq,KEYCODE_D);
              else
                seq_set_1(&defaults->seq,KEYCODE_SPACE);
              break;
            case IPT_JOYSTICKRIGHT_RIGHT:
              if (ui_mode > 0 && max_players > 1)
                seq_set_1(&defaults->seq,KEYCODE_G);
              else
                seq_set_1(&defaults->seq,KEYCODE_LALT);
              break;
            case IPT_UI_CONFIGURE:
              if (ui_mode > 0 && max_players > 1)
                seq_set_2(&defaults->seq,KEYCODE_1,KEYCODE_2);
              break;
            case IPT_UI_UP:
            case IPT_JOYSTICK_UP:
            case IPT_JOYSTICKLEFT_UP:
              seq_set_1(&defaults->seq,KEYCODE_UP);
              break;
            case IPT_UI_DOWN:
            case IPT_JOYSTICK_DOWN:
            case IPT_JOYSTICKLEFT_DOWN:
              seq_set_1(&defaults->seq,KEYCODE_DOWN);
              break;
            case IPT_UI_LEFT:
            case IPT_JOYSTICK_LEFT:
            case IPT_JOYSTICKLEFT_LEFT:
              seq_set_1(&defaults->seq,KEYCODE_LEFT);
              break;
            case IPT_UI_RIGHT:
            case IPT_JOYSTICK_RIGHT:
            case IPT_JOYSTICKLEFT_RIGHT:
              seq_set_1(&defaults->seq,KEYCODE_RIGHT);
              break;
            case IPT_UI_SELECT:
            case IPT_BUTTON1:
            case IPT_PEDAL:
              seq_set_1(&defaults->seq,KEYCODE_LCONTROL);
              break;
            case IPT_BUTTON2:
              seq_set_1(&defaults->seq,KEYCODE_LALT);
              break;
            case IPT_BUTTON3:
              seq_set_1(&defaults->seq,KEYCODE_SPACE);
              break;
            case IPT_BUTTON4:
              seq_set_1(&defaults->seq,KEYCODE_LSHIFT);
              break;
            case IPT_BUTTON5:
              seq_set_1(&defaults->seq,KEYCODE_Z);
              break;
            case IPT_BUTTON6:
              seq_set_1(&defaults->seq,KEYCODE_X);
              break;
            case IPT_UI_PAUSE:
              seq_set_1(&defaults->seq,KEYCODE_PAUSE);
              break;

            /* ----- */
            /* CLEAR */
            /* ----- */

            case IPT_START1:
            case IPT_COIN1:
            case IPT_SERVICE1:
              if (!maple_controller_dev[0])
                seq_set_0(&defaults->seq);
              break;

            case IPT_START2:
            case IPT_COIN2:
            case IPT_SERVICE2:
              if (!maple_controller_dev[1])
                seq_set_0(&defaults->seq);
              break;

            case IPT_START3:
            case IPT_COIN3:
            case IPT_SERVICE3:
              if (!maple_controller_dev[2])
                seq_set_0(&defaults->seq);
              break;

            case IPT_START4:
            case IPT_COIN4:
            case IPT_SERVICE4:
              if (!maple_controller_dev[3])
                seq_set_0(&defaults->seq);
              break;

          #if 0 //OVERCLOCK_STUFF
            case IPT_UI_PAN_UP:
              seq_set_1(&defaults->seq,KEYCODE_5);
              break;
            case IPT_UI_PAN_DOWN:
              seq_set_1(&defaults->seq,KEYCODE_T);
              break;
            case IPT_UI_PAN_LEFT:
              seq_set_1(&defaults->seq,KEYCODE_0);
              break;
            case IPT_UI_PAN_RIGHT:
              seq_set_1(&defaults->seq,KEYCODE_O);
              break;
          #else
            case IPT_UI_PAN_UP:
            case IPT_UI_PAN_DOWN:
            case IPT_UI_PAN_LEFT:
            case IPT_UI_PAN_RIGHT:
          #endif
            case IPT_UI_ON_SCREEN_DISPLAY:
            case IPT_UI_SHOW_GFX:
            case IPT_UI_FRAMESKIP_DEC:
            case IPT_UI_FRAMESKIP_INC:
            case IPT_UI_THROTTLE:
            case IPT_UI_SNAPSHOT:
            case IPT_UI_SHOW_PROFILER:
            case IPT_UI_TOGGLE_DEBUG:
            case IPT_TILT:
            case IPT_UI_TOGGLE_CHEAT:
            case IPT_BUTTON7:
            case IPT_BUTTON8:
            case IPT_BUTTON9:
            case IPT_BUTTON10:
              //printf("Cleared: %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              seq_set_0(&defaults->seq);
              break;

            /* ----- */
            /* LEAVE */
            /* ----- */
                                          
            case IPT_UI_SAVE_STATE:       /* ------------------------------ */
            case IPT_UI_LOAD_STATE:       /* These entries overlap player 1 */
            case IPT_UI_ADD_CHEAT:        /* controls - if they didn't they */
            case IPT_UI_DELETE_CHEAT:     /* would be toast now...          */
            case IPT_UI_SAVE_CHEAT:       /* ------------------------------ */
            case IPT_UI_WATCH_VALUE:
            default:
              //printf("Left:    %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              break;
          }
        }
        break;

        case 1: /* PLAYER2 */
        {
          switch(defaults->type & ~IPF_PLAYERMASK)
          {
            /* --- */
            /* SET */
            /* --- */

            case IPT_JOYSTICK_UP:
            case IPT_JOYSTICKLEFT_UP:
              seq_set_1(&defaults->seq,KEYCODE_R);
              break;
            case IPT_JOYSTICK_DOWN:
            case IPT_JOYSTICKLEFT_DOWN:
              seq_set_1(&defaults->seq,KEYCODE_F);
              break;
            case IPT_JOYSTICK_LEFT:
            case IPT_JOYSTICKLEFT_LEFT:
              seq_set_1(&defaults->seq,KEYCODE_D);
              break;
            case IPT_JOYSTICK_RIGHT:
            case IPT_JOYSTICKLEFT_RIGHT:
              seq_set_1(&defaults->seq,KEYCODE_G);
              break;
            case IPT_BUTTON1:
            case IPT_PEDAL:
              seq_set_1(&defaults->seq,KEYCODE_A);
              break;
            case IPT_BUTTON2:
              seq_set_1(&defaults->seq,KEYCODE_S);
              break;
            case IPT_BUTTON3:
              seq_set_1(&defaults->seq,KEYCODE_Q);
              break;
            case IPT_BUTTON4:
              seq_set_1(&defaults->seq,KEYCODE_W);
              break;

            /* ----- */
            /* CLEAR */
            /* ----- */

            case IPT_BUTTON7:
            case IPT_BUTTON8:
            case IPT_BUTTON9:
            case IPT_BUTTON10:
              //printf("Cleared: %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              seq_set_0(&defaults->seq);
              break;

            /* ----- */
            /* LEAVE */
            /* ----- */

            default:
              //printf("Left:    %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              break;
          }
        }
        break;

        case 2: /* PLAYER3 */
        {
          switch(defaults->type & ~IPF_PLAYERMASK)
          {
            /* --- */
            /* SET */
            /* --- */

            case IPT_JOYSTICK_UP:
            case IPT_JOYSTICKLEFT_UP:
              seq_set_1(&defaults->seq,KEYCODE_I);
              break;
            case IPT_JOYSTICK_DOWN:
            case IPT_JOYSTICKLEFT_DOWN:
              seq_set_1(&defaults->seq,KEYCODE_K);
              break;
            case IPT_JOYSTICK_LEFT:
            case IPT_JOYSTICKLEFT_LEFT:
              seq_set_1(&defaults->seq,KEYCODE_J);
              break;
            case IPT_JOYSTICK_RIGHT:
            case IPT_JOYSTICKLEFT_RIGHT:
              seq_set_1(&defaults->seq,KEYCODE_L);
              break;
            case IPT_BUTTON1:
            case IPT_PEDAL:
              seq_set_1(&defaults->seq,KEYCODE_RCONTROL);
              break;
            case IPT_BUTTON2:
              seq_set_1(&defaults->seq,KEYCODE_RSHIFT);
              break;
            case IPT_BUTTON3:
              seq_set_1(&defaults->seq,KEYCODE_STOP);
              break;
            case IPT_BUTTON4:
              seq_set_1(&defaults->seq,KEYCODE_SLASH);
              break;

            /* ----- */
            /* CLEAR */
            /* ----- */

            case IPT_BUTTON7:
            case IPT_BUTTON8:
            case IPT_BUTTON9:
            case IPT_BUTTON10:
              //printf("Cleared: %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              seq_set_0(&defaults->seq);
              break;

            /* ----- */
            /* LEAVE */
            /* ----- */

            default:
              //printf("Left:    %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              break;
          }
        }
        break;

        case 3: /* PLAYER4 */
        {
          switch(defaults->type & ~IPF_PLAYERMASK)
          {
            /* --- */
            /* SET */
            /* --- */

            case IPT_JOYSTICK_UP:
            case IPT_JOYSTICKLEFT_UP:
              seq_set_1(&defaults->seq,KEYCODE_COLON);
              break;
            case IPT_JOYSTICK_DOWN:
            case IPT_JOYSTICKLEFT_DOWN:
              seq_set_1(&defaults->seq,KEYCODE_QUOTE);
              break;
            case IPT_JOYSTICK_LEFT:
            case IPT_JOYSTICKLEFT_LEFT:
              seq_set_1(&defaults->seq,KEYCODE_BACKSLASH);
              break;
            case IPT_JOYSTICK_RIGHT:
            case IPT_JOYSTICKLEFT_RIGHT:
              seq_set_1(&defaults->seq,KEYCODE_COMMA);
              break;
            case IPT_BUTTON1:
              seq_set_1(&defaults->seq,KEYCODE_F1);
              break;
            case IPT_BUTTON2:
              seq_set_1(&defaults->seq,KEYCODE_F2);
              break;
            case IPT_BUTTON3:
              seq_set_1(&defaults->seq,KEYCODE_OPENBRACE);
              break;
            case IPT_BUTTON4:
              seq_set_1(&defaults->seq,KEYCODE_CLOSEBRACE);
              break;

            /* ----- */
            /* CLEAR */
            /* ----- */

            case IPT_BUTTON7:
            case IPT_BUTTON8:
            case IPT_BUTTON9:
            case IPT_BUTTON10:
              //printf("Cleared: %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              seq_set_0(&defaults->seq);
              break;

            /* ----- */
            /* LEAVE */
            /* ----- */

            default:
              //printf("Left:    %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
              break;
          }
        }
        break;

        default:
          //printf("Unknown: %d '%s' player %d\n",defaults->type,defaults->name,player>>16);
          break;
      }
    }

		defaults++;
	}
}
